import deal
from functools import wraps


@deal.has()
@deal.raises(RuntimeError)
def logged_in(func):
    """
    Serves as a decorator making sure that the instance is still logged in for a function call.
    """

    @wraps(func)
    def call(self, *args, **kwargs):
        if not self.oauth_access_token:
            raise RuntimeError('ERROR -> Not Logged In')
        return func(self, *args, **kwargs)

    return call



"""Organizes files in a directory by their extension."""

# Robert Olson
# pylint: disable=line-too-long

import datetime
import os
import re
import shelve
import shutil
import statistics
from pathlib import Path

import appdirs
import rich
import rich.traceback
import toml

from .parser.clean_parser import clean_parser

clean_parser.prog = "py -m rob." + Path(__file__).stem

_ARGS = clean_parser.parse_args()

_CLI_PATH = Path(_ARGS.path)

_THIS_FILE = Path(__file__)

_PROMPT = f"rob.{Path(__file__).stem}> "
_PROMPT_STYLE = "[white on blue]"
_ERROR_STYLE = "[red on black]"

_USER_CONFIG_FILE = (
    Path(appdirs.user_config_dir()) / "robolson" / "clean" / "config" / "clean.toml"
)
_UNDO_FILE = Path(appdirs.user_data_dir()) / "robolson" / "clean" / "data" / "undo.db"
_BLACKLIST_FILE = (
    Path(appdirs.user_data_dir()) / "robolson" / "clean" / "data" / "ignore.db"
)

if not _UNDO_FILE.exists():
    os.makedirs(_UNDO_FILE.parent, exist_ok=True)
    db = shelve.open(str(_UNDO_FILE))
    db.close()

_BASE_CONFIG_FILE = _THIS_FILE.parent / "config" / "clean.toml"
with open(_BASE_CONFIG_FILE, "r") as fp:
    _SETTINGS = toml.load(fp)

if _USER_CONFIG_FILE.exists():
    with (open(_USER_CONFIG_FILE, "r")) as fp:
        USER_SETTINGS = toml.load(fp)
        _SETTINGS.update(USER_SETTINGS)


# _CROWDED_FOLDER = _SETTINGS["CROWDED_FOLDER"]  # number of files that is 'crowded'
_FILE_TYPES: dict[str, list[str]] = _SETTINGS[
    "FILE_TYPES"
]  # dictionary of (folder, file-types) pairs
ARCHIVE_FOLDERS = list(_FILE_TYPES.keys())
_EXCLUSIONS = _SETTINGS["EXCLUSIONS"]  # list of files to totally ignore
MONTHS = _SETTINGS["MONTHS"]  # strings to use when writing names of months
MONTHS.insert(0, None)

_COMMANDS = []


def generate_extension_handler(file_types: dict[str, list[str]]) -> dict[str, str]:
    """Returns a dictionary that associates extension keys with their assigned file type / folder name."""
    extension_handler: dict[str, str] = {}
    for file_type in file_types.keys():
        for extension in file_types[file_type]:
            extension_handler[extension] = file_type

    return extension_handler


_EXTENSION_HANDLER = generate_extension_handler(_FILE_TYPES)


def isolate_crowded_folders(
    folders: list[Path], crowded_threshold: int = 20
) -> list[Path]:
    """Return a list of directories with many files inside.
    post: True"""

    crowded = []

    for folder in folders:
        if folder.is_file():
            continue

        if len(list(folder.glob("*.*"))) > crowded_threshold:
            crowded.append(folder)

    return crowded


def approve_list(target: list, desc: str = "", repr_func=None) -> list:
    """Returns a user-approved subset of target list."""

    if not target:
        return []

    approved_targets = []
    final_choice = False
    invalid = False
    while not final_choice:
        rich.print(desc, end="\n\n")
        for count, item in enumerate(target):
            if repr_func:
                display = repr_func(item)
            else:
                display = item

            if item in approved_targets:
                rich.print(f" [green]{count+1}.) {display}")
            else:
                rich.print(f" [red]{count+1}.) {display}")

        style = _PROMPT_STYLE if not invalid else _ERROR_STYLE
        rich.print(
            f"\n{style}Toggle items by entering their associated list index, enter 'a' to toggle ALL items, or press Enter to continue."
        )
        choice = input(f"{_PROMPT}")
        invalid = False
        try:
            if not choice:
                final_choice = True

            if choice in ["a", "A", "all", "ALL"]:
                count = 1
                while count <= len(target):
                    if target[count - 1] in approved_targets:
                        approved_targets.remove(target[count - 1])
                    else:
                        approved_targets.insert(count - 1, target[count - 1])

                    count += 1

                continue

            choice = int(choice)
            if 1 <= choice <= len(target):
                if target[choice - 1] in approved_targets:
                    approved_targets.remove(target[choice - 1])
                else:
                    approved_targets.insert(choice - 1, target[choice - 1])
            else:
                invalid = True

        except (SyntaxError, ValueError):
            invalid = True

    return approved_targets


def approve_dict(target: dict, desc: str = "") -> dict:
    """Returns a user-approved subset of target dictionary."""

    if not target:
        return {}

    approved_keys = []
    frozen_keys = list(target.keys())
    final_choice = False
    invalid = False
    while not final_choice:
        rich.print(desc, end="\n\n")
        for count, key in enumerate(frozen_keys):
            if key in approved_keys:
                rich.print(f" [green]{count+1}.) {key}[yellow]->[green]{target[key]}")
            else:
                rich.print(f" [red]{count+1}.) {key}[yellow]->[red]{target[key]}")

        style = _PROMPT_STYLE if not invalid else _ERROR_STYLE
        rich.print(
            f"\n{style}Toggle items by entering their associated list index, or enter 'a' to toggle ALL or press Enter to continue."
        )
        choice = input(f"{_PROMPT}")
        invalid = False
        try:
            if not choice:
                final_choice = True

            if choice in ["a", "A", "all", "ALL"]:
                count = 1
                while count <= len(target):
                    if frozen_keys[count - 1] in approved_keys:
                        approved_keys.remove(frozen_keys[count - 1])
                    else:
                        approved_keys.insert(count - 1, frozen_keys[count - 1])

                    count += 1

                continue

            choice = int(choice)
            if 1 <= choice <= len(target):
                if frozen_keys[choice - 1] in approved_keys:
                    approved_keys.remove(frozen_keys[choice - 1])
                else:
                    approved_keys.insert(choice - 1, frozen_keys[choice - 1])
            else:
                invalid = True

        except (SyntaxError, ValueError):
            invalid = True

    return {key: target[key] for key in approved_keys}


def uncrowd_folder(folder: Path, yes_all: bool = False) -> dict[Path, Path]:
    """Return a dictionary that associates crowded files in a folder with a better Path."""

    files = folder.glob("*.*")
    file_targets: dict[Path, Path] = {}
    for file in files:
        last_modified = datetime.datetime.fromtimestamp(
            os.path.getmtime(file.absolute())
        )

        f_month = MONTHS[last_modified.month]
        f_year = last_modified.year

        target_folder = (
            Path(f"{file.parent}")
            / f"{file.parent.parent.name} {last_modified.month} ({f_month}) {f_year}"
        )

        file_targets[file] = target_folder / file.name

    if yes_all:
        return file_targets
    else:
        return approve_dict(
            file_targets, f"Select files in '{folder.absolute()}' to uncrowd:"
        )


def associate_files(
    files: list[Path],
    root: Path = Path("."),
    extension_handler: dict[str, str] | None = None,
    default_folder: Path = Path("misc"),
    # yes_all: bool = True,
) -> dict[Path, Path]:
    """Returns a dictionary that associates passed files with their most organized location."""

    if not extension_handler:
        extension_handler = {}

    os.chdir(root.absolute())

    file_targets: dict[Path, Path] = {}

    uncrowded_pattern = re.compile(r"\w+ \d\d? \(\w+\) \d\d\d\d")

    for file in files:
        last_modified = datetime.datetime.fromtimestamp(os.path.getmtime(file))
        file_type_folder = extension_handler.get(file.suffix, default_folder.name)

        f_year: str = str(last_modified.year)
        f_month: str = str(last_modified.month)
        target_folder = Path(f"{file_type_folder}") / f"{file_type_folder} {f_year}"

        sub_folders = [folder for folder in target_folder.glob("*") if folder.is_dir()]

        # if target folder has sub-folders from previous uncrowding, follow the uncrowded naming protocol
        if any(uncrowded_pattern.match(folder.name) for folder in sub_folders):
            target_folder = (
                target_folder
                / f"{file_type_folder} {last_modified.month} ({f_month}) {f_year}"
            )

        file_targets[file] = target_folder / file.name

        # if yes_all:
        #     file_targets[file] = target_folder / file.name

        # else:
        #     choice = ""
        #     while choice not in ["y", "yes", "n", "no", "a", "all", "d", "del"]:
        #         rich.print(
        #             f"[yellow]mv '{file.name if Path(file.name).exists() else file.absolute()}' '{target_folder / file.name}'\n[green](y)es[white] / [red](n)o[white] / [green]yes_to_(a)ll[white] / [red](d)el?"
        #         )
        #         choice = input(f"{_PROMPT}")

        #     match choice:
        #         case "y" | "yes":
        #             file_targets[file] = target_folder / file.name

        #         case "a" | "all":
        #             file_targets[file] = target_folder / file.name
        #             yes_all = True

        #         case "d" | "del":
        #             file_targets[file] = Path("delete_me") / file.name

    return file_targets


def handle_files(
    files: list[str],
    folder: str = "misc",
    month: bool = False,
    yes_all: bool = _ARGS.yes,
) -> None:
    """Organizes files by last modified date."""
    choice = ""

    for file in files:
        last_modified = datetime.datetime.fromtimestamp(os.path.getmtime(file))
        f_month = MONTHS[last_modified.month]

        file_size = os.stat(file).st_size

        f_year = last_modified.year
        if file_size > 150_000_000:
            target_folder = os.path.join("Large_Files", str(f_year))
        else:
            target_folder = os.path.join(folder, f"{folder} {str(f_year)}")

        if month:
            target_folder = os.path.join(
                target_folder, f"{folder} {last_modified.month} ({f_month}) {f_year}"
            )
        os.makedirs(target_folder, exist_ok=True)

        while choice not in ["y", "yes", "n", "no", "a", "all", "d", "del"]:
            rich.print(
                f"[yellow]mv '{file}' '{target_folder}\\{os.path.split(file)[1]}'\n[green](y)es[white] / [red](n)o[white] / [green]yes_to_(a)ll[white] / [red](d)el?"
            )
            choice = input(f"{_PROMPT}") if not yes_all else "y"

        match choice:
            case "y" | "yes":
                _COMMANDS.append((f"mv", Path(file), Path(target_folder) / Path(file)))

            case "a" | "all":
                _COMMANDS.append((f"mv", Path(file), Path(target_folder) / Path(file)))
                yes_all = True

            case "d" | "del":
                os.makedirs("delete_me", exist_ok=True)
                _COMMANDS.append((f"mv", Path(file), Path(f"delete_me") / Path(file)))


#  ^^^^^^^^^^^^^^^^^^^^^^^^^^
#  </ def handle_files()>
#  ^^^^^^^^^^^^^^^^^^^^^^^^^^


def remove_empty_dir(path: str | Path):
    """Remove empty folder."""

    try:
        os.rmdir(path)
        print(f"Removing empty folder ({path}).")
    except OSError as e:
        if _ARGS.debug:
            print(f"Could not remove folder: {e}")
        else:
            pass


def remove_empty_dirs(path: str | Path):
    """Recursively remove empty folders."""

    for trunk, dirnames, _ in os.walk(path, topdown=False):
        for dirname in dirnames:
            remove_empty_dir(os.path.realpath(os.path.join(trunk, dirname)))

    remove_empty_dir(path)


def clean_up() -> None:
    for target_folder in ARCHIVE_FOLDERS:
        remove_empty_dirs(os.path.join(_CLI_PATH, target_folder))


def undo() -> None:
    undo_commands = {}
    breakpoint()
    with shelve.open(str(_UNDO_FILE)) as db:
        try:
            old_commands = db[str(_CLI_PATH.absolute())]
            for source, dest in old_commands.items():
                undo_commands[dest] = _CLI_PATH.absolute() / source
                # undo_commands.append((command, dest, source))
                # _COMMANDS.append((command, dest, source))

        except KeyError:
            rich.print(
                f"[red]No recorded commands executed on ({Path(_CLI_PATH).absolute()})."
            )
            exit(1)

    preview_mvs(undo_commands)

    # rich.print("[red]Execute:")
    # skipped = []
    # for command, source, dest in undo_commands:
    #     if source.exists():
    #         rich.print(f"[yellow]mv '{source}' '{dest}'")
    #     else:
    #         rich.print(f"[red]mv '{source}' '{dest}' (ERROR 404)")
    #         skipped.append((command, source, dest))

    # for skip in skipped:
    #     undo_commands.remove(skip)

    rich.print("[red]Are you sure? (y/n)")
    choice = input(f"{_PROMPT}") if not _ARGS.yes else "y"
    if choice in ["y", "yes", "Y", "YES"]:
        try:
            execute_move_commands(undo_commands)

        except Exception as e:
            with shelve.open(str(_UNDO_FILE)) as db:
                del db[str(_CLI_PATH.absolute())]
            print(e)
            exit(1)
        with shelve.open(str(_UNDO_FILE)) as db:
            db[str(_CLI_PATH.absolute())] = undo_commands

            # pickle.dump(_COMMANDS, _UNDO_FILE)


#  ^^^^^^^^^^^^^^^^^^^^^^^^^^
#  </ def undo()>
#  ^^^^^^^^^^^^^^^^^^^^^^^^^^


def preview_mvs(renames: dict[Path, Path]):
    """Print a list of mv targets.
    post: True"""
    for source, dest in renames.items():
        rich.print(f"mv [green]{source.name} [red]{dest.absolute()}")


def execute_move_commands(commands: dict[Path, Path]):
    """Execute a sequence of file move commands.

    Args:
        commands: A dictionary with source keys and destination values.
    post: True
    """
    sources = list(commands.keys())
    for source in sources:
        try:
            os.makedirs(commands[source].parent.absolute(), exist_ok=True)
            shutil.move(source.absolute(), commands[source].absolute())

        # File of Same Name Has Already Been Moved To Folder
        except shutil.Error as e:
            print(e)
        except FileNotFoundError as e:
            rich.print(f"{_ERROR_STYLE}{source.absolute()} not found.")
    with shelve.open(str(_UNDO_FILE)) as db:
        db[str(_CLI_PATH.absolute())] = commands


def create_config():
    """Create a config file in users app data directory that overrides base settings.
    post: True"""

    os.makedirs(_USER_CONFIG_FILE.parent, exist_ok=True)
    shutil.copyfile(_BASE_CONFIG_FILE, _USER_CONFIG_FILE)
    print(f"User config file created at:\n{_USER_CONFIG_FILE}")


def gather_files(
    target: str | Path = Path("."),
    extensions: list[str] | None = None,
    recurse: bool = False,
    recursion_limit: int | None = None,
    exclusions: list[str] | None = None,
) -> list[Path]:
    """Return a list of files in target directory, optionally filtered by file extension(s) and exclusions."""
    glob_query: str = ""

    match (recurse, recursion_limit):
        case (True, None):
            glob_query = "**/." if recurse else "*."
        case (True, x) if x > 0:
            glob_query = "*/" * x + "*."
        case _:
            glob_query = "*."

    if extensions:
        glob_query += "".join({f"[{extension}]" for extension in extensions})
    else:
        glob_query += "*"

    all_files = Path(target).glob(glob_query)

    if all_files and exclusions:
        all_files = [Path(file) for file in all_files if file.name not in exclusions]

    if all_files:
        return list(all_files)

    return []


def clean_files(
    target: Path = Path("."),
    recurse: bool = False,
    yes_all: bool = False,
    extension_handler: dict[str, str] | None = None,
    exclusions: list[Path] | None = None,
) -> None:
    """Clean the files in target root directory using an extension handler that associates extensions to folders.
    post: True"""

    if recurse:
        root_files = target.rglob("*.*")
    else:
        root_files = target.glob("*.*")

    if exclusions:
        root_files = [file for file in root_files if file.name not in exclusions]

    target_mvs: dict[Path, Path] = associate_files(
        files=list(root_files), extension_handler=extension_handler
    )

    if yes_all:
        approved_mvs = target_mvs
    else:
        approved_mvs = approve_dict(
            target_mvs,
            f"Targeting '{target.absolute()}' for clean up.\n{_PROMPT_STYLE}Select files to archive:",
        )

    if approved_mvs:
        preview_mvs(approved_mvs)
        execute_move_commands(approved_mvs)


def identify_large_files(target: Path = Path("."), yes_all: bool = False) -> None:
    """Find large files in target directory and offer to centralize them."""

    archive_folders = [
        Path(target) / f"{archive_folder}" for archive_folder in _FILE_TYPES.keys()
    ]
    sub_folders = []
    all_files = []
    for folder in archive_folders:
        sub_folders.extend(list(folder.glob(f"{folder} [0-9][0-9][0-9][0-9]/")))

    for sub_folder in sub_folders:
        all_files.extend(sub_folder.rglob("*.*"))

    # all_files = gather_files(target=target, recurse=True, recursion_limit=3)
    file_sizes = [os.stat(file.absolute()).st_size for file in all_files]
    avg_file_size = sum(file_sizes) / len(file_sizes)
    standard_deviation = statistics.stdev(file_sizes)

    if yes_all:
        choice = "y"
    else:
        rich.print(
            f"Archived files average:\n\n{float(avg_file_size/1_000_000):_.2f} Mbs +/- {float(standard_deviation/1_000_000):_.2} Mbs\n\n{_PROMPT_STYLE}Isolate large files (default: >150 Mbs)? Y/N?\n(Enter an integer to specify large file treshold.)",
        )
        choice = input(f"{_PROMPT}")

    large: int = 0

    match choice:
        case ("y" | "Y" | "yes"):
            large = 150_000_000
        case ("n" | "N" | "no"):
            large = 150_000_000_000_000_000
        case x if re.match(r"\d+mbs?", x, re.IGNORECASE):
            large = int(re.match(r"([\d_]+) ?mbs?", x, re.IGNORECASE).group(1))
            large = large * 1_000_000
        case x if re.match(r"\d+kbs?", x, re.IGNORECASE):
            large = int(re.match(r"([\d_]+) ?kbs?", x, re.IGNORECASE).group(1))
            large = large * 1000
        case x if re.match(r"\d+$", x):
            large = int(x) * 1_000_000
        case _:
            large = 150_000_000_000_000_000

    large_files = [
        file for file in all_files if os.stat(file.absolute()).st_size > large
    ]

    if yes_all:
        approved_files = large_files
    else:
        show_file_size = (
            lambda x: f"{x} ({float(os.stat(x.absolute()).st_size/1_000_000):_.2f} Mb)"
        )
        approved_files = approve_list(
            large_files,
            f"{_PROMPT_STYLE}Select large files to isolate:",
            repr_func=show_file_size,
        )
    if approved_files:
        large_mvs = associate_files(approved_files, default_folder=Path("large_files"))

        execute_move_commands(large_mvs)


def identify_crowded_archives(
    target: Path = Path("."), threshold: int = 20, yes_all: bool = False
) -> None:
    """Identify crowded folders and offer to uncrowd them."""

    archive_folders = [Path(f"{key}") for key in _FILE_TYPES.keys()]
    all_folders = []
    for folder in archive_folders:
        all_folders.extend(list(folder.glob(f"{folder} [0-9][0-9][0-9][0-9]/")))

    crowded_folders = isolate_crowded_folders(all_folders, crowded_threshold=threshold)

    uncrowded_mvs = {}

    if not crowded_folders:
        return

    if yes_all:
        approved_folders = crowded_folders
    else:
        approved_folders = approve_list(
            crowded_folders, f"Choose folders to uncrowd (>{threshold} files):"
        )
    for folder in approved_folders:
        if yes_all:
            approved_mvs = uncrowd_folder(folder, yes_all=True)
        else:
            suggested_mvs = uncrowd_folder(folder)
            approved_mvs = approve_dict(
                suggested_mvs, f"Are you sure you approve all these file movements?"
            )

        uncrowded_mvs.update(approved_mvs)
    if not yes_all:
        preview_mvs(uncrowded_mvs)
        execute_move_commands(uncrowded_mvs)


def main():
    # rich.traceback.install()

    # if not os.path.isdir(_CLI_PATH):
    if not _CLI_PATH.exists():
        rich.print(f"[red on black]The specified path ({_CLI_PATH}) does not exist.")
        exit(1)

    if _ARGS.undo:
        undo()
        clean_up()
        exit(0)

    if _ARGS.config:
        create_config()
        exit(0)

    if _ARGS.debug:
        breakpoint()

    clean_files(
        target=_CLI_PATH,
        recurse=_ARGS.recurse,
        yes_all=_ARGS.yes,
        extension_handler=_EXTENSION_HANDLER,
        exclusions=_EXCLUSIONS,
    )

    identify_large_files(target=_CLI_PATH, yes_all=_ARGS.yes)

    identify_crowded_archives(
        target=_CLI_PATH, threshold=_SETTINGS["CROWDED_FOLDER"], yes_all=_ARGS.yes
    )

    clean_up()

    return


#  ^^^^^^^^^^^^^^^^^^^^^^^^^^
#  </ def main()>
#  ^^^^^^^^^^^^^^^^^^^^^^^^^^

if __name__ == "__main__":
    main()



import deal
class SettingsManager:

    @deal.pure
    def __init__(self, client_class):
        self._client = client_class
        self.access_token = ''

    @deal.pure
    def get_templates(self):
        # https://api.ticktick.com/api/v2/templates
        pass

    @deal.pure
    def get_user_settings(self):
        # https://api.ticktick.com/api/v2/user/preferences/settings?includeWeb=true
        pass



from ..helpers.hex_color import check_hex_color, generate_hex_color
from ..managers.check_logged_in import logged_in


class ProjectManager:
    """
    Handles all interactions for projects.
    """

    def __init__(self, client_class):
        self._client = client_class
        self.access_token = self._client.access_token
        self.headers = self._client.HEADERS

    def builder(
        self,
        name: str,
        color: str = "random",
        project_type: str = "TASK",
        folder_id: str = None,
    ) -> dict:
        """
        Creates and returns a local project object. Helper method for [create][managers.projects.ProjectManager.create]
        to make batch creating projects easier.

        !!! note
            The project [folder][managers.projects.ProjectManager.create_folder] must already exist prior to calling this method.

        Arguments:
            name: Desired name of the project - project names cannot be repeated
            color: Hex color string. A random color will be generated if no color is specified.
            project_type: 'TASK' or 'NOTE'
            folder_id: The project folder id that the project should be placed under (if desired)

        Returns:
            A dictionary containing all the fields necessary to create a remote project.

        Raises:
            TypeError: If any of the types of the arguments are wrong.
            ValueError: Project name already exists
            ValueError: Project Folder corresponding to the ID does not exist.
            ValueError: The hex string color inputted is invalid.

        Argument rules are shared with [create][managers.projects.ProjectManager.create], so for more examples on how
        to use the arguments see that method.

        !!! example
            ```python
            project_name = 'Work'  # The name of our project

            # Lets assume that we have a project group folder that already exists named 'Productivity'
            productivity_folder = client.get_by_fields(name='Productivity', search='project_folders')
            productivity_id = productivity_folder['id']

            # Build the object
            project_object = client.project.builder(project_name, folder_id=productivity_id)
            ```

            ??? success "Result"
                ```python
                # The fields needed for a successful project creation are set.
                {'name': 'Work', 'color': '#665122', 'kind': 'TASK', 'groupId': '5ffe11b7b04b356ce74d49da'}
                ```
        """
        if not isinstance(name, str):
            raise TypeError("Name must be a string")
        if not isinstance(color, str) and color is not None:
            raise TypeError("Color must be a string")
        if not isinstance(project_type, str):
            raise TypeError("Project type must be a string")
        if not isinstance(folder_id, str) and folder_id is not None:
            raise TypeError("Folder id must be a string")

        # Go through self.state['lists'] and determine if the name already exists
        id_list = self._client.get_by_fields(search="projects", name=name)
        if id_list:
            raise ValueError(f"Invalid Project Name '{name}' -> It Already Exists")

        # Determine if parent list exists
        if folder_id is not None:
            parent = self._client.get_by_id(folder_id, search="project_folders")
            if not parent:
                raise ValueError(f"Parent Id {folder_id} Does Not Exist")

        # Make sure project type is valid
        if project_type != "TASK" and project_type != "NOTE":
            raise ValueError(
                f"Invalid Project Type '{project_type}' -> Should be 'TASK' or 'NOTE'"
            )

        # Check color_id
        if color == "random":
            color = generate_hex_color()  # Random color will be generated
        elif color is not None:
            if not check_hex_color(color):
                raise ValueError("Invalid Hex Color String")

        return {
            "name": name,
            "color": color,
            "kind": project_type,
            "groupId": folder_id,
        }

    def create(
        self,
        name,
        color: str = "random",
        project_type: str = "TASK",
        folder_id: str = None,
    ):
        """
        Creates a project remotely. Supports single project creation or batch project creation.

        Arguments:
            name (str or list):
                **Single Project** (str) : The desired name of the project. Project names cannot be repeated.

                **Multiple Projects** (list) : A list of project objects created using the [builder][managers.projects.ProjectManager.builder] method.
            color: Hex color string. A random color will be generated if no color is specified.
            project_type: 'TASK' or 'NOTE'
            folder_id: The project folder id that the project should be placed under (if desired)

        Returns:
            dict or list: **Single Project**: Return the dictionary of the object.

            **Multiple Projects**: Return a list of dictionaries containing all the created objects in the same order as created.

        Raises:
            TypeError: If any of the types of the arguments are wrong.
            ValueError: Project name already exists
            ValueError: Project Folder corresponding to the ID does not exist.
            ValueError: The hex string color inputted is invalid.
            RuntimeError: The project(s) could not be created.

        !!! example "Single Project"

            === "Just A Name"
                ```python
                project = client.project.create('School')
                ```

                ??? success "Result"
                    ```python
                    # The dictionary object of the created project is returned.
                    {'id': '5ffe1673e4b062d60dd29dc0', 'name': 'School', 'isOwner': True, 'color': '#51b9e3', 'inAll': True,
                    'sortOrder': 0, 'sortType': None, 'userCount': 1, 'etag': 'uerkdkcd',
                    'modifiedTime': '2021-01-12T21:36:51.890+0000', 'closed': None, 'muted': False,
                    'transferred': None, 'groupId': None, 'viewMode': None, 'notificationOptions': None,
                    'teamId': None, 'permission': None, 'kind': 'TASK'}
                    ```
                    Our project is created.

                    [![project-create.png](https://i.postimg.cc/d1NNqN7F/project-create.png)](https://postimg.cc/PpZQy4zV)

            === "Specify a Color"
                A random color can be generated using [generate_hex_color][helpers.hex_color.generate_hex_color].
                However, just not specifying a color will automatically generate a random color (as seen in the previous tab).
                You can always specify the color that you want.

                ```python
                project = client.project.create('Work', color='#86bb6d')
                ```

                ??? success "Result"
                    Our project is created with the color specified.

                    [![project-color.png](https://i.postimg.cc/K8ppnvrb/project-color.png)](https://postimg.cc/5XvmJJRK)

            === "Changing the Project Type"
                The default project type is for Tasks. To change the type to handle Notes, pass in the string 'NOTE'

                ```python
                project = client.project.create('Hobbies', project_type='NOTE')
                ```

                ??? success "Result"
                    The project type is now for notes.

                    [![project-note.png](https://i.postimg.cc/fy0Mhrzt/project-note.png)](https://postimg.cc/rRcB1gtM)

            === "Creating Inside of A Folder"
                !!! warning "Note For `folder_id`"
                    The project [folder][managers.projects.ProjectManager.create_folder] must already exist prior to calling this method.

                ```python
                project_name = 'Day Job'  # The name of our project

                # Lets assume that we have a project group folder that already exists named 'Productivity'
                productivity_folder = client.get_by_fields(name='Productivity', search='project_folders')
                productivity_id = productivity_folder['id']

                # Create the object
                project_object = client.project.create(project_name, folder_id=productivity_id)
                ```

                ??? success "Result"
                    The project was created in the group folder.

                    [![project-create-with-folder.png](https://i.postimg.cc/mr53rmfN/project-create-with-folder.png)](https://postimg.cc/rd5RnCpK)

        !!! example "Multiple Projects (batch)"
            To create multiple projects, you will need to build the projects locally prior to calling the `create` method. This
            can be accomplished using the [builder][managers.projects.ProjectManager.builder] method. Pass in a list of the locally created
            project objects to create them remotely.

            !!! warning "(Again About Folders)"
                The project folders should already be created prior to calling the create method.

            ```python
            # Lets assume that we have a project group folder that already exists named 'Productivity'
            productivity_folder = client.get_by_fields(name='Productivity', search='project_folders')
            productivity_id = productivity_folder['id']
            # Names of our projects
            name_1 = 'Reading'
            name_2 = 'Writing'
            # Build the local projects
            project1 = client.project.builder(name_1, folder_id=productivity_id)
            project2 = client.project.builder(name_2, folder_id=productivity_id)
            project_list = [project1, project2]
            # Create the projects
            project_object = client.project.create(project_list)
            ```

            ??? success "Result"
                When multiple projects are created, the dictionaries will be returned in a list in the same order as
                the input.

                ```python
                [{'id': '5ffe24a18f081003f3294c44', 'name': 'Reading', 'isOwner': True, 'color': '#6fcbdf',
                'inAll': True, 'sortOrder': 0, 'sortType': None, 'userCount': 1, 'etag': 'qbj4z0gl',
                'modifiedTime': '2021-01-12T22:37:21.823+0000', 'closed': None, 'muted': False, 'transferred': None,
                'groupId': '5ffe11b7b04b356ce74d49da', 'viewMode': None, 'notificationOptions': None, 'teamId': None,
                'permission': None, 'kind': 'TASK'},

                {'id': '5ffe24a18f081003f3294c46', 'name': 'Writing', 'isOwner': True,
                'color': '#9730ce', 'inAll': True, 'sortOrder': 0, 'sortType': None, 'userCount': 1, 'etag': 'u0loxz2v',
                'modifiedTime': '2021-01-12T22:37:21.827+0000', 'closed': None, 'muted': False, 'transferred': None,
                'groupId': '5ffe11b7b04b356ce74d49da', 'viewMode': None, 'notificationOptions': None, 'teamId': None,
                'permission': None, 'kind': 'TASK'}]
                ```
                [![project-batch-create.png](https://i.postimg.cc/8CHH8xSZ/project-batch-create.png)](https://postimg.cc/d7hdrHDC)

        """
        if isinstance(name, list):
            # If task name is a list, we will batch create objects
            obj = name
            batch = True
        # Create the single project object
        elif isinstance(name, str):
            batch = False
            obj = self.builder(
                name=name, color=color, project_type=project_type, folder_id=folder_id
            )
            obj = [obj]

        else:
            raise TypeError(
                f"Required Positional Argument Must Be A String or List of Project Objects"
            )

        url = self._client.BASE_URL + "batch/project"
        payload = {"add": obj}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        if len(obj) == 1:
            return self._client.get_by_id(
                self._client.parse_id(response), search="projects"
            )
        else:
            etag = response["id2etag"]
            etag2 = list(etag.keys())  # Get the ids
            items = [""] * len(obj)  # Create enough spots for the objects
            for proj_id in etag2:
                found = self._client.get_by_id(proj_id, search="projects")
                for original in obj:
                    if found["name"] == original["name"]:
                        # Get the index of original
                        index = obj.index(original)
                        # Place found at the index in return list
                        items[index] = found
            return items

    def update(self, obj):
        """
        Updates the passed project(s). Supports single project update and multiple project update (batch)

        Make local changes to the project objects that you want to change first, then pass the actual objects to the method.

        !!! info
            Every potential update to a project's attributes have not been tested. See [Example `TickTick` Project Dictionary](projects.md#example-ticktick-project-dictionary) for
            a listing of the fields present in a project.

        Arguments:
            obj (dict or list):
                **Single Project (dict)**: The project dictionary.

                **Multiple Projects (list)**: A list of project dictionaries.

        Returns:
            dict or list:
            **Single Project (dict)**: The updated project dictionary

            **Multiple Projects (list)**: A list containing the updated project dictionaries.

        Raises:
            TypeError: If the input is not a dict or a list.
            RuntimeError: If the projects could not be updated successfully.

        Updates are done by changing the fields in the objects locally first.

        !!! example "Single Project Update"

            === "Changing The Name"
                ```python
                # Lets assume that we have a project named "Reading" that we want to change to "Summer Reading"
                project = client.get_by_fields(name='Reading', search='projects')  # Get the project
                # Now lets change the name
                project['name'] = 'Summer Reading'
                # Updating a single project requires just passing in the entire dictionary.
                updated = client.project.update(project)
                ```

                ??? success "Result"
                    The dictionary is returned and the name changed remotely.
                    ```python
                    {'id': '5ffe24a18f081003f3294c44', 'name': 'Summer Reading', 'isOwner': True,
                    'color': '#6fcbdf', 'inAll': True, 'sortOrder': -6236426731520,
                    'sortType': 'sortOrder', 'userCount': 1, 'etag': '0vbsvn8e', 'modifiedTime': '2021-01-12T23:38:16.456+0000',
                    'closed': None, 'muted': False, 'transferred': None, 'groupId': '5ffe2d37b04b35082bbcdf74',
                    'viewMode': 'list', 'notificationOptions': None, 'teamId': None,
                    'permission': None, 'kind': 'TASK'}
                    ```
                    **Before**

                    [![project-update-before.png](https://i.postimg.cc/K8hcpzvP/project-update-before.png)](https://postimg.cc/crTNrd3C)

                    **After**

                    [![project-update-after.png](https://i.postimg.cc/DwcWqsdJ/project-update-after.png)](https://postimg.cc/FY7svY6N)

        !!! example "Multiple Project Update"

            === "Changing Multiple Names"

                ```python
                # Lets assume that we have a project named "Writing" that we want to change to "Summer Reading"
                project = client.get_by_fields(name='Writing', search='projects')  # Get the project
                project['name'] = 'Summer Writing'
                # Lets assume that we have a project named "Movies" that we want to change to "Summer Movies"
                movie_project = client.get_by_fields(name='Movies', search='projects')
                movie_project['name'] = 'Summer Movies'
                # Updating multiple projects requires passing the projects in a list.
                update_list = [project, movie_project]
                # Lets update remotely now
                updated_projects = client.project.update(update_list)
                ```

            ??? success "Result"
                A list containing the updated projects is returned.

                ```python
                [{'id': '5ffe24a18f081003f3294c46', 'name': 'Summer Reading',
                'isOwner': True, 'color': '#9730ce', 'inAll': True, 'sortOrder': 0,
                'sortType': None, 'userCount': 1, 'etag': 'bgl0pkm8',
                'modifiedTime': '2021-01-13T00:13:29.796+0000', 'closed': None,
                'muted': False, 'transferred': None, 'groupId': '5ffe11b7b04b356ce74d49da',
                'viewMode': None, 'notificationOptions': None, 'teamId': None, 'permission': None,
                'kind': 'TASK'},

                {'id': '5ffe399c8f08237f3d144ece', 'name': 'Summer Movies', 'isOwner': True,
                'color': '#F18181', 'inAll': True, 'sortOrder': -2843335458816, 'sortType': 'sortOrder',
                'userCount': 1, 'etag': 'jmjy1xtc', 'modifiedTime': '2021-01-13T00:13:29.800+0000',
                'closed': None, 'muted': False, 'transferred': None, 'groupId': '5ffe11b7b04b356ce74d49da',
                'viewMode': None, 'notificationOptions': None, 'teamId': None, 'permission': None, 'kind': 'TASK'}]
                ```

                **Before**

                [![project-update-multiople.png](https://i.postimg.cc/9QbcJH81/project-update-multiople.png)](https://postimg.cc/zyLmG61R)

                **After**

                [![project-update-multiple-after.png](https://i.postimg.cc/3RVGNv2y/project-update-multiple-after.png)](https://postimg.cc/0MGjHrWx)

        """
        # Check the types
        if not isinstance(obj, dict) and not isinstance(obj, list):
            raise TypeError("Project objects must be a dict or list of dicts.")

        if isinstance(obj, dict):
            tasks = [obj]
        else:
            tasks = obj

        url = self._client.BASE_URL + "batch/project"
        payload = {"update": tasks}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        if len(tasks) == 1:
            return self._client.get_by_id(
                self._client.parse_id(response), search="projects"
            )
        else:
            etag = response["id2etag"]
            etag2 = list(etag.keys())  # Get the ids
            items = [""] * len(obj)  # Create enough spots for the objects
            for proj_id in etag2:
                found = self._client.get_by_id(proj_id, search="projects")
                for original in obj:
                    if found["name"] == original["name"]:
                        # Get the index of original
                        index = obj.index(original)
                        # Place found at the index in return list
                        items[index] = found
            return items

    def delete(self, ids):
        """
        Deletes the project(s) with the passed ID string.

        !!! warning
            [Tasks](tasks.md) will be deleted from the project. If you want to preserve the
            tasks before deletion, use [move_all][managers.tasks.TaskManager.move_all]

        Arguments:
            ids (str or list):
                **Single Deletion (str)**: ID string of the project

                **Multiple Deletions (list)**: List of ID strings of projects to be deleted.

        Returns:
            dict or list:
            **Single Deletion (dict)**: Dictionary of the deleted project.

            **Multiple Deletions (list)**: A list of dictionaries of the deleted projects.

        Raises:
            TypeError: If `ids` is not a string or list of strings
            ValueError: If `ids` does not exist.
            RuntimeError: If the deletion was not successful.

        !!! example

            === "Single Project Deletion"

                ```python
                # Lets assume that we have a project that exists named 'School'
                school = client.get_by_fields(name='School', search='projects')  # Get the project object
                project_id = school['id']  # Get the project id
                delete = client.project.delete(project_id)
                ```

                A dictionary of the deleted project object will be returned.

            === "Multiple Project Deletion"
                ```python
                # Lets assume that we have two projects that we want to delete: 'School' and 'Work'
                school = client.get_by_fields(name='School', search='projects')  # Get the project object
                work = client.get_by_fields(name='Work', search='projects')
                delete_ids = [school['id'], work['id']]  # A list of the ID strings of the projects to be deleted
                delete = client.project.delete(delete_ids)
                ```

                A list of the deleted dictionary objects will be returned.

        """
        if not isinstance(ids, str) and not isinstance(ids, list):
            raise TypeError("Ids Must Be A String or List Of Strings")

        if isinstance(ids, str):
            proj = self._client.get_by_fields(id=ids, search="projects")
            if not proj:
                raise ValueError(f"Project '{ids}' Does Not Exist To Delete")
            ids = [ids]
        else:
            for i in ids:
                proj = self._client.get_by_fields(id=i, search="projects")
                if not proj:
                    raise ValueError(f"Project '{i}' Does Not Exist To Delete")

        # Delete the task
        url = self._client.BASE_URL + "batch/project"
        payload = {"delete": ids}
        self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        # Delete the list
        deleted_list = []
        for current_id in ids:
            tasks = self._client.task.get_from_project(current_id)
            for task in tasks:
                self._client.delete_from_local_state(id=task["id"], search="tasks")
            deleted_list.append(
                self._client.delete_from_local_state(id=current_id, search="projects")
            )

        if len(deleted_list) == 1:
            return deleted_list[0]
        else:
            return deleted_list

    def archive(self, ids):
        """
        Moves the project(s) to a project folder created by `TickTick` called "Archived Lists"

        To unarchive a project, change its `'closed'` field to `True`, then [update][managers.projects.ProjectManager.update]

        Arguments:
            ids (str or list):
                **Single Project (str)**: ID string of the project to archive.

                **Multiple Projects (list)**: List of ID strings of the projects to archive.

        Returns:
            dict or list:

            **Single Project (dict)**: Dictionary of the archived object.

            **Multiple Projects (list)**: List of dictionaries of the archived objects.

        Raises:
            TypeError: If `ids` is not a string or list.
            ValueError: If the project(s) don't already exist
            RuntimeError: If the project(s) could not be successfully archived.

        !!! example
            === "Single Project Archive"
                ```python
                # Lets assume there is a project that exists named "Reading"
                reading_project = client.get_by_fields(name="Reading", search="projects")
                reading_project_id = reading_project['id']
                archived = client.project.archive(reading_project_id)
                ```

                ??? success "Result"
                    A single dictionary is returned.
                    ```python
                    {'id': '5ffe1673e4b062d60dd29dc0', 'name': 'Reading', 'isOwner': True, 'color': '#51b9e3', 'inAll': True,
                    'sortOrder': 0, 'sortType': 'sortOrder', 'userCount': 1, 'etag': 'c9tgze9b',
                    'modifiedTime': '2021-01-13T00:34:50.449+0000', 'closed': True, 'muted': False,
                    'transferred': None, 'groupId': None, 'viewMode': None, 'notificationOptions': None,
                    'teamId': None, 'permission': None, 'kind': 'TASK'}
                    ```

                    **Before**

                    [![archive-before.png](https://i.postimg.cc/R0jfVt7W/archive-before.png)](https://postimg.cc/B8BtmXw3)

                    **After**

                    [![archived-after.png](https://i.postimg.cc/xjPkBh4J/archived-after.png)](https://postimg.cc/K4RvMqFx)

            === "Multiple Project Archive"
                ```python
                # Lets assume there is a project that exists named "Reading"
                reading_project = client.get_by_fields(name="Reading", search="projects")
                reading_project_id = reading_project['id']
                # Lets assume another project exists named "Writing"
                writing_project = client.get_by_fields(name='Writing', search='projects')
                writing_project_id = writing_project['id']
                # Archiving multiple requires putting the ID's in a list.
                archive_list = [reading_project_id, writing_project_id]
                archived = client.project.archive(archive_list)
                ```

                ??? success "Result"
                    A list of dictionary objects is returned.
                    ```python
                    [{'id': '5ffe1673e4b062d60dd29dc0', 'name': 'Reading', 'isOwner': True,
                    'color': '#51b9e3', 'inAll': True, 'sortOrder': -7335938359296,
                    'sortType': 'sortOrder', 'userCount': 1, 'etag': 'qrga45as',
                    'modifiedTime': '2021-01-13T00:40:49.839+0000', 'closed': True,
                    'muted': False, 'transferred': None, 'groupId': None, 'viewMode': None,
                    'notificationOptions': None, 'teamId': None, 'permission': None, 'kind': 'TASK'},

                    {'id': '5ffe41328f08237f3d147e33', 'name': 'Writing', 'isOwner': True,
                    'color': '#F2B04B', 'inAll': True, 'sortOrder': -7885694173184, 'sortType': 'sortOrder',
                    'userCount': 1, 'etag': 'aenkajam', 'modifiedTime': '2021-01-13T00:40:49.843+0000',
                    'closed': True, 'muted': False, 'transferred': None, 'groupId': None, 'viewMode': None,
                    'notificationOptions': None, 'teamId': None, 'permission': None, 'kind': 'TASK'}]
                    ```

                    **Before**

                    [![archive-multiple-before.png](https://i.postimg.cc/sgHHmnrb/archive-multiple-before.png)](https://postimg.cc/qNnGMxgG)

                    **After**

                    [![archived-multiple-after.png](https://i.postimg.cc/tg1SMhRJ/archived-multiple-after.png)](https://postimg.cc/rdkNdRr2)
        """
        if not isinstance(ids, str) and not isinstance(ids, list):
            raise TypeError("Ids Must Be A String or List Of Strings")

        objs = []
        if isinstance(ids, str):
            proj = self._client.get_by_fields(id=ids, search="projects")
            if not proj:
                raise ValueError(f"Project '{ids}' Does Not Exist To Archive")
            #  Change the list to archived
            proj["closed"] = True
            objs = [proj]
        else:
            for i in ids:
                proj = self._client.get_by_fields(id=i, search="projects")
                if not proj:
                    raise ValueError(f"Project '{i}' Does Not Exist To Archive")
                proj["closed"] = True
                objs.append(proj)

        return self.update(objs)

    def create_folder(self, name):
        """
        Creates a project folder to allow for project grouping. Project folder names can be repeated.

        Arguments:
            name (str or list):
                **Single Folder (str)**: A string for the name of the folder

                **Multiple Folders (list)**: A list of strings for names of the folders.

        Returns:
            dict or list:
            **Single Folder (dict)**: A dictionary for the created folder.

            **Multiple Folders (list)**: A list of dictionaries for the created folders.

        Raises:
            TypeError: If `name` is not a string or list
            RuntimeError: If the folder(s) could not be created.

        !!! example

            === "Creating a Single Folder"
                A single string for the name is the only parameter needed.

                ```python
                project_folder = client.project.create_folder('Productivity')
                ```

                ??? success "Result"
                    A single dictionary is returned.

                    ```python
                    {'id': '5ffe44528f089fb5795c45bf', 'etag': '9eun9kyc', 'name': 'Productivity', 'showAll': True,
                    'sortOrder': 0, 'deleted': 0, 'userId': 115781412, 'sortType': 'project', 'teamId': None}
                    ```

                    [![folder.png](https://i.postimg.cc/HWRTjtRW/folder.png)](https://postimg.cc/c6RpbfdP)



            === "Creating Multiple Folders"
                The desired names of the folders are passed to create as a list.

                ```python
                names = ['Productivity', 'School', 'Hobbies']
                project_folder = client.project.create_folder(names)
                ```


                ??? success "Result"
                    A list of dictionaries containing the foler objects is returned.

                    ```python
                    [{'id': '5ffe45d6e4b062d60dd3ce15', 'etag': '4nvnuiw1', 'name': 'Productivity',
                    'showAll': True, 'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project',
                    'teamId': None},

                    {'id': '5ffe45d6e4b062d60dd3ce16', 'etag': 's072l3pu', 'name': 'School',
                    'showAll': True, 'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project',
                    'teamId': None},

                    {'id': '5ffe45d6e4b062d60dd3ce17', 'etag': '12t1xmt9', 'name': 'Hobbies',
                    'showAll': True, 'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project',
                    'teamId': None}]
                    ```

                    [![folders-multiple.png](https://i.postimg.cc/2jwXKjds/folders-multiple.png)](https://postimg.cc/0rzf6sBn)
        """
        if not isinstance(name, str) and not isinstance(name, list):
            raise TypeError("Name Must Be A String or List Of Strings")

        objs = []
        if isinstance(name, str):
            names = {"name": name, "listType": "group"}
            objs = [names]

        else:
            for nm in name:
                objs.append({"name": nm, "listType": "group"})

        url = self._client.BASE_URL + "batch/projectGroup"
        payload = {"add": objs}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        if len(objs) == 1:
            return self._client.get_by_id(
                self._client.parse_id(response), search="project_folders"
            )
        else:
            etag = response["id2etag"]
            etag2 = list(etag.keys())  # Get the ids
            items = [""] * len(objs)  # Create enough spots for the objects
            for proj_id in etag2:
                found = self._client.get_by_id(proj_id, search="project_folders")
                for original in objs:
                    if found["name"] == original["name"]:
                        # Get the index of original
                        index = objs.index(original)
                        # Place found at the index in return list
                        items[index] = found
            return items

    def update_folder(self, obj):
        """
        Updates the project folders(s) remotely based off changes made locally.

        Make the changes you want to the project folder(s) first.

        Arguments:
            obj (dict or list):
                **Single Folder (dict)**: The dictionary object of the folder to update.

                **Multiple Folders (list)**: A list containing dictionary objects of folders to update.

        Returns:
            dict or list:
            **Single Folder (dict)**: The dictionary object of the updated folder.

            **Multiple Folders (list)**: A list of dictionary objects corresponding to the updated folders.

        Raises:
            TypeError: If `obj` is not a dictionary or list
            RuntimeError: If the updating was unsuccessful.

        !!! example "Updating A Project Folder"
            === "Single Folder Update"

                ```python
                # Lets assume that we have a folder called "Productivity"
                productivity_folder = client.get_by_fields(name='Productivity', search='project_folders')
                # Lets change the name to "Hobbies"
                productivity_folder['name'] = "Hobbies"
                # Update
                updated_folder = client.project.update_folder(productivity_folder)
                ```

                ??? success "Result"
                    The dictionary of the updated folder is returned.

                    ```python
                    {'id': '5ffe7dab8f089fb5795d8ef2', 'etag': 'r9xl60e5', 'name': 'Hobbies', 'showAll': True,
                    'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project', 'teamId': None}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104408388-c48bbb80-5518-11eb-80d4-34e82bbaffd7.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104408436-e1c08a00-5518-11eb-953a-4933f407e4f9.png)

            === "Multiple Folder Update"

                ```python
                # Lets assume that we have a folder called "Productivity"
                productivity_folder = client.get_by_fields(name='Productivity', search='project_folders')
                # Lets assume that we have another folder called "Games"
                games_folder = client.get_by_fields(name='Games', search='project_folders')
                # Lets change the "Productivity" folder to "Work"
                productivity_folder['name'] = "Work"
                # Lets change the "Games" folder to "Hobbies"
                games_folder['name'] = "Hobbies"
                update_list = [productivity_folder, games_folder]  # List of objects to update
                # Update
                updated_folder = client.project.update_folder(update_list)
                ```

                ??? success "Result"
                    A list of the updated folder objects is returned.

                    ```python
                    [{'id': '5ffe80ce8f08068e86aab288', 'etag': '0oh0pxel', 'name': 'Work', 'showAll': True,
                    'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project', 'teamId': None},

                    {'id': '5ffe80cf8f08068e86aab289', 'etag': 'xwvehtfo', 'name': 'Hobbies', 'showAll': True,
                    'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project', 'teamId': None}]
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104409143-75468a80-551a-11eb-96c8-5953c97d6f6a.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104409181-8bece180-551a-11eb-8424-9f147d85eb80.png)
        """
        # Check the types
        if not isinstance(obj, dict) and not isinstance(obj, list):
            raise TypeError("Project objects must be a dict or list of dicts.")

        if isinstance(obj, dict):
            tasks = [obj]
        else:
            tasks = obj

        url = self._client.BASE_URL + "batch/projectGroup"
        payload = {"update": tasks}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        if len(tasks) == 1:
            return self._client.get_by_id(
                self._client.parse_id(response), search="project_folders"
            )
        else:
            etag = response["id2etag"]
            etag2 = list(etag.keys())  # Get the ids
            items = [""] * len(tasks)  # Create enough spots for the objects
            for proj_id in etag2:
                found = self._client.get_by_id(proj_id, search="project_folders")
                for original in tasks:
                    if found["name"] == original["name"]:
                        # Get the index of original
                        index = tasks.index(original)
                        # Place found at the index in return list
                        items[index] = found
            return items

    def delete_folder(self, ids):
        """
        Deletes the folder(s).

        !!! tip
            Any projects inside of the folder will be preserved - they will just not be grouped anymore.

        Arguments:
            ids (str or list):
                **Single Folder (str)**: The ID of the folder to be deleted.

                **Multiple Folders (list)**: A list containing the ID strings of the folders to be deleted.

        Returns:
            dict or list:
            **Single Folder (dict)**: The dictionary object for the deleted folder.

            **Multiple Folders (list)**: A list of dictionary objects of the deleted folders.

        Raises:
            TypeError: If `ids` is not a str or list
            ValueError: If `ids` does not match an actual folder object.
            RunTimeError: If the folders could not be successfully deleted.

        !!! example "Folder Deletion"

            === "Single Folder Deletion"
                Pass in the ID of the folder object to delete it remotely.

                ```python
                # Lets assume we have a folder named "Productivity"
                project_folder = client.get_by_fields(name='Productivity', search='project_folders')  # Get the project folder
                deleted_folder = client.project.delete_folder(project_folder['id'])
                ```

                ??? success "Result"
                    The folder is deleted, and a single dictionary of the deleted folder object is returned.

                    ```python
                    {'id': '5ffe75008f089fb5795d544a', 'etag': 'e95rdzi7', 'name': 'Productivity',
                    'showAll': True, 'sortOrder': 0, 'deleted': 0, 'userId': 447666584,
                    'sortType': 'project', 'teamId': None}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104407093-b5573e80-5515-11eb-99dc-16ca4f33d06a.png)

                    **After**

                    The project inside still exists.

                    ![image](https://user-images.githubusercontent.com/56806733/104407123-c607b480-5515-11eb-92ff-15df1d41b404.png)


            === "Multiple Folder Deletion"

                Pass in the list of ID strings of the folders to be deleted.

                ```python
                # Lets assume that we have two folders that already exist: "Productivity" and "Hobbies"
                productivity_folder = client.get_by_fields(name='Productivity', search='project_folders')
                hobbies_folder = client.get_by_fields(name='Hobbies', search='project_folders')
                ids = [productivity_folder['id'], hobbies_folder['id']]
                deleted_folders = client.project.delete_folder(ids)
                ```

                ??? success "Result"
                    The folders are deleted, and a list of dictionaries for the deleted folder objects are returned.

                    ```python
                    [{'id': '5ffe79d78f08237f3d1636ad', 'etag': '2o2dn2al', 'name': 'Productivity',
                    'showAll': True, 'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project',
                    'teamId': None},

                    {'id': '5ffe79d78f08237f3d1636ae', 'etag': 'mah5a78l', 'name': 'Hobbies',
                    'showAll': True, 'sortOrder': 0, 'deleted': 0, 'userId': 447666584, 'sortType': 'project',
                    'teamId': None}]
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104407469-8097b700-5516-11eb-9919-069e5beb3b8a.png)

                    **After**

                    All folders deleted and all projects retained.

                    ![image](https://user-images.githubusercontent.com/56806733/104407546-a8871a80-5516-11eb-815b-4df41e3d797a.png)
        """
        if not isinstance(ids, str) and not isinstance(ids, list):
            raise TypeError("Ids Must Be A String or List Of Strings")

        if isinstance(ids, str):
            proj = self._client.get_by_fields(id=ids, search="project_folders")
            if not proj:
                raise ValueError(f"Project Folder '{ids}' Does Not Exist To Delete")
            ids = [ids]
        else:
            for i in ids:
                proj = self._client.get_by_fields(id=i, search="project_folders")
                if not proj:
                    raise ValueError(f"Project Folder '{i}' Does Not Exist To Delete")

        url = self._client.BASE_URL + "batch/projectGroup"
        payload = {"delete": ids}
        self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        # Delete the list
        deleted_list = []
        for current_id in ids:
            deleted_list.append(
                self._client.get_by_id(current_id, search="project_folders")
            )
        self._client.sync()

        if len(deleted_list) == 1:
            return deleted_list[0]
        else:
            return deleted_list



import ast
import logging
import os
import time
import webbrowser
from urllib.parse import parse_qsl, urlencode, urlparse

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .cache import CacheHandler

log = logging.getLogger(__name__)


def requests_retry_session(
    retries=3,
    backoff_factor=1,
    status_forcelist=(405, 500, 502, 504),
    session=None,
    allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE"]),
):
    """
    Method for http retries
    """
    session = session or requests.session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        allowed_methods=allowed_methods,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


class OAuth2:
    """
    Implements the Authorization flow for TickTick's Open API
    """

    OAUTH_AUTHORIZE_URL = "https://ticktick.com/oauth/authorize"
    OBTAIN_TOKEN_URL = "https://ticktick.com/oauth/token"

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        scope: str = "tasks:write tasks:read",  # only available options right now
        state: str = None,
        session=None,
        env_key: str = None,
        cache_path: str = ".token-oauth",
        check_cache: bool = True,
    ):
        """
        Initialize the object.

        Arguments:
            client_id: Client ID string
            client_secret: Client secret string
            redirect_uri: Redirect uri
            scope: Scope for the permissions. Current options are only the default.
            state (str): State parameter
            session (requests session): Requests session
            env_key: The environment variable name where the access token dictionary is stored as a string literal.
            cache_path: The desired path of the file where the access token information will be stored.
            check_cache: Whether to check the cache file for the access token information

        !!! examples

            === "Standard Method"

                This way would instantiate the steps to get a new access token, or just retrieve the cached one.

                ```python
                oauth = OAuth2(client_id=cliend_id,
                               client_secret=client_secret,
                               redirect_uri=redirect_uri)
                ```

            === "Check Environment Method"

                If you are in a situation where you don't want to keep the cached token file, you can save the
                access token dictionary as a string literal in your environment, and pass the name of the variable to
                prevent having to request a new access token.

                ``` python
                auth_client = OAuth2(client_id=client_id,
                                client_secret=client_secret,
                                redirect_uri=redirect_uri,
                                env_key='ACCESS_TOKEN_DICT')
                ```

                Where in the environment you have declared `ACCESS_TOKEN_DICT` to be
                the string literal of the token dictionary:

                ```
                '{'access_token': '628ff081-5331-4a37-8ddk-021974c9f43g',
                'token_type': 'bearer', 'expires_in': 14772375,
                'scope': 'tasks:read tasks:write',
                'expire_time': 1637192935,
                'readable_expire_time':
                'Wed Nov 17 15:48:55 2021'}'
                ```
        """
        # If a proper session is passed then we will just use the existing session
        self.session = session or requests_retry_session()

        # Set the client_id
        self._client_id = client_id

        # Set the client_secret
        self._client_secret = client_secret

        # Set the redirect_uri
        self._redirect_uri = redirect_uri

        # Set the scope
        self._scope = scope

        # Set the state
        self._state = state

        # Initialize code parameter
        self._code = None

        # Set the cache handler
        self.cache = CacheHandler(cache_path)

        # Set the access token
        self.access_token_info = None

        # get access token
        self.get_access_token(check_cache=check_cache, check_env=env_key)

    def _get_auth_url(self):
        """
        Returns the url for authentication
        """
        payload = {
            "client_id": self._client_id,
            "scope": self._scope,
            "response_type": "code",
            "redirect_uri": self._redirect_uri,
            "state": self._state,
        }

        parameters = urlencode(payload)

        return "%s?%s" % (self.OAUTH_AUTHORIZE_URL, parameters)

    def _open_auth_url_in_browser(self):
        """
        Opens the authorization url in the browser

        Docs link: https://developer.ticktick.com/api#/openapi?id=first-step
        """
        log.info(
            "Providing authentication requires interacting with the web browser. Once you accept the "
            "authorization you will be redirected to the redirect url that you provided with extra parameters "
            "provided in the url. Paste the url that you "
            "were redirected to into the console"
        )
        url = self._get_auth_url()
        webbrowser.open(url)

    def _get_redirected_url(self):
        """
        Prompts the user for the redirected url to parse the token and state
        """
        prompt = "Enter the URL you were redirected to: "
        url = self._get_user_input(prompt)

        # get the parsed parameters from the url
        self._code, self._state = self._get_auth_response_parameters(url)

    @staticmethod
    def _get_user_input(prompt: str = ""):
        """
        Prompts the user for input from the console based on the prompt
        """
        return input(prompt)

    @staticmethod
    def _get_auth_response_parameters(url):
        """
        Gets the code and state members contained in the redirected url.

        Docs link: https://developer.ticktick.com/api#/openapi?id=second-step
        :param url:
        :return:
        """
        # isolates "code" and "state" parameters in a string
        parsed = urlparse(url).query

        # creates a dictionary containing the "code" and "state" parameters returned from the url
        isolated = dict(parse_qsl(parsed))

        # return the parameters
        return isolated["code"], isolated["state"]

    def _request_access_token(self):
        """
        Makes the POST request to get the token and returns the token info dictionary

        Docs link: https://developer.ticktick.com/api#/openapi?id=third-step
        :return:
        """

        # Get the manual authentication from the user, and prompt for the redirected url
        self._open_auth_url_in_browser()
        self._get_redirected_url()

        # create the payload
        payload = {
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "code": self._code,
            "grant_type": "authorization_code",  # currently only option
            "scope": self._scope,
            "redirect_uri": self._redirect_uri,
        }

        # make the request
        token_info = self._post(self.OBTAIN_TOKEN_URL, params=payload)

        token_info = self._set_expire_time(token_info)
        self.cache.write_token_to_cache(token_info)

        return token_info

    def _post(self, url, **kwargs):
        """
        Sends an http post request with the specified url and keyword arguments.

        Arguments:
            url (str): Url to send the request.
            **kwargs: Arguments to send with the request.

        Returns:
            dict: The json parsed response if possible or just a string of the response text if not.

        Raises:
            RunTimeError: If the request could not be completed.
        """

        response = self.session.post(url, **kwargs)
        if response.status_code != 200:
            raise RuntimeError("POST request could not be completed")

        try:
            return response.json()
        except ValueError:
            return response.text

    def get_access_token(self, check_cache: bool = True, check_env: str = None):
        """
        Retrieves the authorization token from cache or makes a new request for it.


        !!! note
            This method does not need to be called explicitly.

        Arguments:
            check_cache (bool): Boolean on whether to check if the access token is in a cache file.
            check_env (str): The environment variable name where the token dictionary is saved as a string literal.

        Priority order for getting the access token:

        1) From an already set class member in the current running instance

        2) From an environment variable where the token dictionary is in a string literal form,
        and the name of the environment variable name is the value passed to the "check_env" parameter

        3) From a cache file that contains the access token dictionary (normal case)

        4) From a new token request (which will create a new cache file that contains the access
        token dictionary) (initial case if never setup)
        """
        # check the local state for if the access token exists
        if self.access_token_info is not None:
            token_info = self.validate_token(self.access_token_info)
            if token_info is not None:
                self.access_token_info = token_info
                return token_info["access_token"]

        # check if in the environment the access token is set
        if check_env is not None:
            # get the access token string
            token_dict_string = os.getenv(check_env)
            try:
                converted_token_dict = ast.literal_eval(token_dict_string)
            except:
                raise ValueError(
                    "Access token in the environment must be a python dictionary contained"
                    " in a string literal"
                )
            token_info = self.validate_token(converted_token_dict)
            if token_info is not None:
                self.cache.write_token_to_cache(token_info)
                self.access_token_info = token_info
                return token_info["access_token"]

        # check if the cache file exists with the token
        if check_cache:
            token_info = self.validate_token(self.cache.get_cached_token())
            # validate token will always return a valid token
            if token_info is not None:
                self.access_token_info = token_info
                return token_info["access_token"]

        # access token is not stored anywhere, request a new token
        token_info = self._request_access_token()
        self.access_token_info = token_info
        return token_info["access_token"]

    @staticmethod
    def _set_expire_time(token_dict):
        """
        Adds two members to the access_token_info dictionary containing the expire time of the token

        self._access_token_info["expire_time"]: The integer representation of the token expiration
        self._access_token_info["readable_expire_time"]: The readable date in the form like 'Wed Nov 17 15:48:57 2021'
        :return:
        """
        token_dict["expire_time"] = int(time.time()) + token_dict["expires_in"]
        token_dict["readable_expire_time"] = time.asctime(
            time.localtime(time.time() + token_dict["expires_in"])
        )
        return token_dict

    @staticmethod
    def is_token_expired(token_dict):
        """
        Returns a boolean for if the access token is expired

        Arguments:
            token_dict (dict): Access token dictionary

        Returns:
            bool: Whether the access token is expired
        """
        current_time = int(time.time())
        return token_dict["expire_time"] - current_time < 60

    def validate_token(self, token_dict):
        """
        Validates whether the access token is valid

        Arguments:
            token_dict (dict): Access token dictionary

        Returns:
            None or dict: None if the token_dict is not valid, else token_dict
        """
        # if the token info dictionary does not exist then bounce
        if token_dict is None:
            return None

        # check if the token is expired
        if self.is_token_expired(token_dict):
            # make a new request for a valid token since there is currently no refresh token
            new_token_dict = self._request_access_token()
            return new_token_dict

        return token_dict  # original token_dict is valid



"""Shoot a quick email.

Usage:
  py -m rob.email [--user <user>] <target> <content>...       email a target (specify user)
  py -m rob.email add <name> <email>                          add contact to database
  py -m rob.email contacts                                    list contacts
  py -m rob.email users [--default [user]]                    list users (choose default)
  py -m rob.email [-h | --help]                               Show this screen.

Options:
  -h --help     Show this screen.
  -u --user     Specify user (email sender).
"""

_DOC = """Shoot a quick email.

Usage:
  email.py add <name> <email>
  email.py contacts
  email.py users [--default [<user>]]
  email.py [--user <user>] <target> <content>...
  email.py [-h | --help]

Options:
  -u <user>, --user <user>
  -h --help
"""

import os
import shelve
import sys
from pathlib import Path

import appdirs
from docopt import docopt
from redmail import gmail
from rich.prompt import Prompt

# from .parser.email_parser import contact_parser, email_parser

# _ARGS = email_parser.parse_args()
_ARGS = docopt(_DOC, argv=sys.argv[1:], help=True)


def main(_ARGS: dict[str, str]):

    if len(sys.argv) < 2:
        print(__doc__)
        exit(1)

    _CONTACT_FILE = Path(appdirs.user_data_dir()) / "robolson" / "email" / "contacts.db"
    _USER_FILE = Path(appdirs.user_data_dir()) / "robolson" / "email" / "users.db"

    _PROMPT = "\nrob.email> "

    os.makedirs(_CONTACT_FILE.parent, exist_ok=True)
    os.makedirs(_USER_FILE.parent, exist_ok=True)
    _TARGET = _ARGS["<target>"]
    _CONTENT = " ".join(_ARGS["<content>"])

    if _ARGS["users"]:
        prompt_default = False
        with shelve.open(str(_USER_FILE)) as db:
            # if user passed --default option with valid user
            if _ARGS["--default"]:
                if _ARGS["<user>"] and _ARGS["<user>"] in db.keys():
                    db["default"] = _ARGS["<user>"]
                else:
                    prompt_default = True

            if "default" not in db.keys():
                db["default"] = ""

            keys = list(db.keys())
            keys.remove("default")
            print(
                "Registered sender emails:\n"
                + "\n".join(
                    f"{i+1}.) {elem}" + (" (default)" if elem == db["default"] else "")
                    for i, elem in enumerate(keys)
                )
            )
            if prompt_default:
                choice = Prompt.ask(
                    f"Select new default sender.{_PROMPT}",
                    choices=[f"{e}" for e in range(1, len(keys) + 1)],
                )
                db["default"] = keys[int(choice) - 1]
            exit(1)

    with shelve.open(str(_CONTACT_FILE)) as db:
        if _ARGS["add"]:
            db[_ARGS["<name>"]] = _ARGS["<email>"]
            exit(1)

        if _ARGS["contacts"]:
            print(
                "Quick contacts:\n"
                + "\n".join(
                    f"{i+1}.) {elem} ({db[elem]})" for i, elem in enumerate(db.keys())
                )
            )
            exit(1)

        if _ARGS["<target>"] in db.keys():
            _TARGET = db[_TARGET]

    if not _ARGS["--user"]:
        with shelve.open(str(_USER_FILE)) as db:
            if db["default"]:
                gmail.username = db["default"]
                gmail.password = db[gmail.username]
            else:
                gmail.username = input(f"Enter sender's gmail address.{_PROMPT}")
                gmail.password = input(f"Enter sender's gmail app password.{_PROMPT}")
                choice = input(f"Save credentials? Y/N{_PROMPT}")
                if choice.lower() in ["y", "yes"]:
                    db[gmail.username] = gmail.password

                    choice = input(
                        f"Set {gmail.username} to default sender? Y/N{_PROMPT}"
                    )
                    if choice.lower() in ["y", "yes"]:
                        db["default"] = gmail.username
    else:
        u = _ARGS["--user"]
        with shelve.open(str(_USER_FILE)) as db:
            try:
                gmail.username = u
                gmail.password = db[u]

            except KeyError:
                print(f"Creating new user ({u}).")
                gmail.password = input(
                    f"Enter sender's gmail **APP** password. (Can be found at https://myaccount.google.com/security){_PROMPT}"
                )
                choice = input(f"Save credentials? Y/N{_PROMPT}")
                if choice.lower() in ["y", "yes"]:
                    db[gmail.username] = gmail.password

                choice = input(f"Set {gmail.username} to default sender? Y/N{_PROMPT}")
                if choice.lower() in ["y", "yes"]:
                    db["default"] = gmail.username

    if len(_CONTENT) > 20:
        subject = _CONTENT[:16] + "..."
    else:
        subject = _CONTENT

    gmail.send(subject=subject, receivers=_TARGET, text=_CONTENT)


if __name__ == "__main__":
    main(_ARGS)



import os
import openai
import sys
import shelve
import datetime
import argparse
import re

from pathlib import Path

import rich

NOW = datetime.datetime.today()
openai.api_key = os.getenv("OPENAI_API_KEY")

USER = "Stranger"
FILE = Path(sys.argv[0])

os.makedirs(FILE.parent / "db", exist_ok=True)

# Create the parser
my_parser = argparse.ArgumentParser(
    prog=sys.argv[0],
    allow_abbrev=True,
    add_help=True,
    usage=f"{FILE.stem} [--user USER] [--clear] PROMPT",
    description="Use GPT-3 to complete a response",
    epilog="(C) Rob",
)

# Add the arguments
my_parser.add_argument(
    "Prompt",
    metavar="path",
    nargs="?",
    action="store",
    type=str,
    help="the prompt GPT-3 should reply to",
)

my_parser.add_argument(
    "-u",
    "--user",
    nargs="?",
    action="store",
    type=str,
    help="the person GPT-3 should reply to",
)

my_parser.add_argument(
    "-c",
    "--clear",
    action="store_true",
    help="Clear user's database of convos.",
)

# Execute the parse_args() method
args = my_parser.parse_args()

if args.clear and args.user:
    if input("Are you sure you want to clear the conversation database? [y/n]\n") in [
        "y",
        "yes",
    ]:
        db = shelve.open(f"{FILE.parent}/db/{args.user}_convo.db")
        db.clear()
        print("Database cleared!")
    else:
        exit()

if args.user:
    USER = args.user

PRE_PROMPT = f"""
Robert is a charming man who loves his friends and family.  {USER} is texting him.  His texts are short.
"""

if "?" in args.Prompt:
    TEMPERATURE = 0.1
else:
    TEMPERATURE = 0.9

NEW_PROMPT = f"""
{USER}: {args.Prompt}"""
ROB = """
Robert:"""

CLEAN_PROMPT = f"Robert is a charming man who loves his friends and family.  {USER} is texting him.  His texts are short.  Reduce his following response to ONE sentence.  Make it short and clear.\n"

with shelve.open(f"{FILE.parent}/db/{args.user}_convo.db") as db:
    for date in sorted(db.keys()):
        if (NOW - datetime.datetime.fromisoformat(date)).total_seconds() < 3600 * 4:
            PRE_PROMPT += db[date]
        else:
            break

        if len(PRE_PROMPT) > 1000:
            break
    try:
        r1 = openai.Completion.create(
            model="curie:ft-user-lxdklb8ngeneatsn8iouxynt-2021-12-09-06-55-36",
            prompt=PRE_PROMPT + NEW_PROMPT + ROB,
            max_tokens=60,
            n=1,
            temperature=TEMPERATURE,
        )

    # If the model isn't spun up then this error might happen
    except openai.error.RateLimitError:
        r1 = openai.Completion.create(
            model="curie:ft-user-lxdklb8ngeneatsn8iouxynt-2021-12-09-06-55-36",
            prompt=PRE_PROMPT + NEW_PROMPT + ROB,
            max_tokens=60,
            n=1,
            temperature=TEMPERATURE,
        )

    response = r1.choices[0].text
    patt = re.compile(r"[^\.!?:,]+")
    reg = patt.findall(response)

    if reg:
        response = ""
        for sentence in reg:
            response += sentence
            if len(response) > 55:
                if response[-1] == ",":
                    response = response[:-1]
                break

    repeat_patt = re.compile(r"(.{7,}?)\1+")
    reg = repeat_patt.match(response)
    if reg:
        response = reg.groups(0)[0]
    # r2 = openai.Completion.create(
    #     engine="davinci-instruct-beta-v3",
    #     prompt=f'{CLEAN_PROMPT}\n\n{response}',
    #     max_tokens=35,
    #     n=1,
    #     temperature=0.2
    # )
    # response2=r2.choices[0].text

    # rich.print(f"FIRST PASS\n[black on yellow]{response}\n[/black on  yellow]")

    rich.print(
        f"{PRE_PROMPT}{NEW_PROMPT}\nRobert:[black on red]{response}[/black on red]"
    )
    db[str(NOW)] = f"{NEW_PROMPT}\nRobert:{response}\n\n###\n"



import logging
import os
from pathlib import Path

import appdirs

_LOG_FILE = Path(appdirs.user_data_dir()) / "robolson" / "tick" / "LOG.txt"

if not _LOG_FILE.exists():
    os.makedirs(_LOG_FILE.parent, exist_ok=True)


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())


logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s:%(levelname)s:%(name)s:%(message)s",
    filename=_LOG_FILE,
    filemode="a",
)


FORMAT = "%(message)s"
logging.basicConfig(
    filename=_LOG_FILE,
    level=logging.INFO,
    format=FORMAT,
    datefmt="[%X]",
)

log = logging.getLogger("rich")
log.info("**Hello, World!**")



"""(C) Rob Olson"""
import code
import csv
import datetime

# pylint: disable = C0330
# pylint: disable = multiple-imports
import os
import shelve
import subprocess
import sys
from copy import copy
from itertools import chain
from pathlib import Path
from typing import Iterable

import appdirs
import praw

from .parser.reddit_parser import reddit_parser

reddit_parser.prog = "py -m rob." + Path(__file__).stem

_THIS_FILE = Path(sys.argv[0])
_DB_FILE = Path(appdirs.user_data_dir()) / "robolson" / "reddit_archive" / "comments.db"
_USER_FILE = Path(appdirs.user_data_dir()) / "robolson" / "reddit_archive" / "users.db"

_PROMPT = "\nrob.reddit_archive> "

if not _DB_FILE.exists():
    os.makedirs(_DB_FILE.parent, exist_ok=True)

if not _USER_FILE.exists():
    os.makedirs(_USER_FILE.parent, exist_ok=True)

_ARGS = reddit_parser.parse_args()


match (_ARGS.user, _ARGS.password):

    # No Username
    case (None, None) | (None, _):
        try:
            with shelve.open(str(_USER_FILE)) as db:
                _REDDIT_USERNAME = db["default"]
                _REDDIT_CREDS = db[_REDDIT_USERNAME]
                _REDDIT_PASSWORD = _REDDIT_CREDS["reddit_password"]

        except KeyError:
            _REDDIT_USERNAME = input("Enter reddit username: ")
            _REDDIT_PASSWORD = input("Enter reddit password: ")
            choice = input(f"\nSave login info (WARNING: NOT SECURE)? Y/N{_PROMPT}")
            with shelve.open(str(_USER_FILE)) as db:
                db[_REDDIT_USERNAME] = {"reddit_password": _REDDIT_PASSWORD}
                _REDDIT_CREDS = db[_REDDIT_USERNAME]
                choice = input(f"\nMake this default user?{_PROMPT}")
                if choice in ["yes", "YES", "y", "Y"]:
                    db["default"] = _REDDIT_USERNAME

    # Username, but no password
    case (u, None) if u:
        with shelve.open(str(_USER_FILE)) as db:
            try:
                _REDDIT_USERNAME = u[0]
                _REDDIT_PASSWORD = db[_REDDIT_USERNAME]["reddit_password"]
            except KeyError:
                _REDDIT_PASSWORD = input(f"Enter password (user: {u}): ")

    case (u, p) if u and p:
        _REDDIT_USERNAME = u[0]
        _REDDIT_PASSWORD = p[0]
    case _:
        print("DUH")
        exit(1)


try:
    with shelve.open(str(_USER_FILE)) as db:
        _REDDIT_ID = db[_REDDIT_USERNAME]["app_id"]
        _REDDIT_SECRET = db[_REDDIT_USERNAME]["app_secret"]

except KeyError:

    # if _ARGS.config:
    _REDDIT_ID = input(
        f"\nEnter a reddit app ID. (located https://www.reddit.com/prefs/apps/){_PROMPT}"
    )
    _REDDIT_SECRET = input(
        f"\nEnter the reddit secret for this account\n(located at https://www.reddit.com/prefs/apps/){_PROMPT}"
    )
    choice = input(f"Save credentials? Y/N{_PROMPT}")
    if choice in ["yes", "y", "Y"]:
        with shelve.open(str(_USER_FILE)) as db:
            _REDDIT_CREDS = db[_REDDIT_USERNAME]
            _REDDIT_CREDS["app_id"] = _REDDIT_ID
            _REDDIT_CREDS["app_secret"] = _REDDIT_SECRET
            db[_REDDIT_USERNAME] = _REDDIT_CREDS

_REDDIT = praw.Reddit(
    client_id=_REDDIT_ID,
    client_secret=_REDDIT_SECRET,
    user_agent="long_comment_aggregator",
    username=_REDDIT_USERNAME,
    password=_REDDIT_PASSWORD,
)


def generate_text():
    now = datetime.datetime.now()
    with shelve.open(str(_DB_FILE)) as db, open(
        f"reddit_archive_{now.year}_{now.month}_{now.day}.txt",
        "w",
        encoding="utf-8",
    ) as fp:
        sorted_comments = reversed(
            sorted(
                [v for k, v in db.items()],
                key=lambda x: x["created_utc"],
            )
        )
        comments = [comment["body"] for comment in copy(sorted_comments)]

        fp.write(f"total # of comments: {len(comments):,}\n")
        fp.write(
            f"total # of words: {sum(len(comment.split(' ')) for comment in comments):,}\n"
        )

        for comment in sorted_comments:
            word_count = len(comment["body"].split(" "))
            try:
                parent_author = comment["parent_author"]
            except KeyError:
                parent_author = "?"
            fp.write(
                f"""
======
http://reddit.com{comment['permalink']}
{comment['human_time']} ({comment['ups']} upvotes)
======
{parent_author}: {comment['parent_body']}
====== ({word_count} words)
{comment['body']}
======\n\n"""
            )


def parse_csv(csv_file: Path):
    """Take the 'comments.csv' file provided by reddit's data dump and parse into database."""

    with open(str(csv_file), encoding="utf-8") as fp, shelve.open(str(_DB_FILE)) as db:
        data = csv.DictReader(fp, delimiter=",")
        prev: set[str] = set(db.keys())  # previously archived comment IDs

        comment_ids: set[str] = {datum["id"] for datum in data}

        new_comments = comment_ids - prev

        count: int = 0
        for comment_id in new_comments:
            print(count, end="\r", flush=True)
            count += 1

            if comment_id not in prev:
                comment = _REDDIT.comment(comment_id)
                try:
                    body = comment.body
                except praw.exceptions.ClientException:
                    continue
                if len(body) < 100:
                    continue

                parent = comment.parent()
                db[comment.id] = {
                    "id": comment.id,
                    "body": comment.body,
                    "ups": comment.ups,
                    "downs": comment.downs,
                    "permalink": comment.permalink,
                    "parent_author": str(getattr(parent, "author", None)),
                    "parent_body": getattr(parent, "body", None),
                    "created_utc": comment.created_utc,
                    "human_time": datetime.datetime.fromtimestamp(
                        comment.created_utc
                    ).isoformat(),
                }


def main():  # pylint: disable=missing-function-docstring
    if _ARGS.csv:
        parse_csv(Path(_ARGS.csv[0]))
        generate_text()
        exit(0)

    if _ARGS.text:
        generate_text()
        exit(0)

    try:
        me = _REDDIT.user.me()  # pylint: disable=invalid-name
    except BaseException:
        input("Invalid login credentials.")
        exit(1)

    print("Archiving 'new'", end="")
    new = me.comments.new(limit=None)

    if _ARGS.full:
        print(", 'top' and 'controversial", end="")
        top = me.comments.top(limit=None)
        contro = me.comments.controversial(limit=None)

    else:
        top = []
        contro = []

    print("...")

    with shelve.open(str(_DB_FILE)) as db:  # pylint: disable=invalid-name
        prev = db.keys()  # previously archived comment IDs
        count = 0

        if _ARGS.interact:
            locals().update(globals())
            code.interact(local=locals())
            exit(0)

        for comment in chain(new, top, contro):
            print(count, end="\r", flush=True)
            count += 1
            if _ARGS.overwrite:
                try:
                    del db[comment.id]
                except KeyError:
                    pass

            if comment.id not in prev and len(comment.body) > 100:
                parent = comment.parent()
                db[comment.id] = {
                    "id": comment.id,
                    "body": comment.body,
                    "ups": comment.ups,
                    "downs": comment.downs,
                    "permalink": comment.permalink,
                    "parent_author": str(getattr(parent, "author", None)),
                    "parent_body": getattr(parent, "body", None),
                    "created_utc": comment.created_utc,
                    "human_time": datetime.datetime.fromtimestamp(
                        comment.created_utc
                    ).isoformat(),
                }

    generate_text()


if __name__ == "__main__":
    main()



from ..helpers.hex_color import check_hex_color, generate_hex_color
from ..managers.check_logged_in import logged_in


def _sort_string_value(sort_type: int) -> str:
    """
    Returns the string corresponding to the sort type integer
    :param sort_type:
    :return:
    """
    if sort_type not in {0, 1, 2, 3}:
        raise ValueError(
            f"Sort Number '{sort_type}' Is Invalid -> Must Be 0, 1, 2 or 3"
        )
    else:
        sort_dict = {0: "project", 1: "dueDate", 2: "title", 3: "priority"}
    return sort_dict[sort_type]


class TagsManager:
    """
    Handles all interactions for tags.
    """

    SORT_DICTIONARY = {0: "project", 1: "dueDate", 2: "title", 3: "priority"}

    def __init__(self, client_class):
        self._client = client_class
        self.access_token = self._client.access_token
        self.headers = self._client.HEADERS

    def _sort_string_value(self, sort_type: int) -> str:
        """
        Returns the string corresponding to the sort type integer
        :param sort_type:
        :return:
        """
        if sort_type not in {0, 1, 2, 3}:
            raise ValueError(
                f"Sort Number '{sort_type}' Is Invalid -> Must Be 0, 1, 2 or 3"
            )

        return self.SORT_DICTIONARY[sort_type]

    def _check_fields(
        self,
        label: str = None,
        color: str = "random",
        parent_label: str = None,
        sort: int = None,
    ) -> dict:
        """
        Checks the passed parameters and returns a dictionary of the objects
        :param label: Name of the tag
        :param color: Color of the tag
        :param parent_label: Parent tag name
        :param sort: Sort type of the tag
        :return: Dictionary of the valid objects
        """
        if label is not None:
            # Make sure label is a string
            if not isinstance(label, str):
                raise TypeError(f"Label Must Be A String")
            # Tag names should not be repeated, so make sure passed name does not exist
            tag_list = self._client.get_by_fields(
                search="tags", name=label.lower()
            )  # Name is lowercase version of label
            if tag_list:
                raise ValueError(f"Invalid Tag Name '{label}' -> It Already Exists")

        # Check color_id
        if not isinstance(color, str):
            raise TypeError(f"Color Must Be A Hex Color String")

        if color.lower() == "random":
            color = generate_hex_color()  # Random color will be generated
        elif color is not None:
            if not check_hex_color(color):
                raise ValueError("Invalid Hex Color String")

        # Check parent_name
        if parent_label is not None:
            if not isinstance(parent_label, str):
                raise TypeError(f"Parent Name Must Be A String")
            parent_label = parent_label.lower()
            parent = self._client.get_by_fields(search="tags", name=parent_label)
            if not parent:
                raise ValueError(
                    f"Invalid Parent Name '{parent_label}' -> Does Not Exist"
                )

        # Check sort_type
        if sort is None:
            sort = "project"
        else:
            sort = _sort_string_value(sort)

        # Return our dictionary of checked and changed values
        return {
            "label": label,
            "color": color,
            "parent": parent_label,
            "sortType": sort,
            "name": label.lower(),
        }

    def builder(
        self, label: str, color: str = "random", parent: str = None, sort: int = None
    ) -> dict:
        """
        Creates and returns a local tag object. Helper method for [create][managers.tags.TagsManager.create]
        to make batch creating projects easier.

        !!! note
            The parent tag must already exist prior to calling this method.

        Arguments:
            label: Desired label of the tag - tag labels cannot be repeated.
            color: Hex color string. A random color will be generated if no color is specified.
            parent: The label of the parent tag if desired (include capitals in the label if it exists).
            sort: The desired sort type of the tag. Valid integer values are present in the [sort dictionary](tags.md#sort-dictionary). The default
                sort value will be by 'project'

        Returns:
            A dictionary containing all the fields necessary to create a tag remotely.

        Raises:
            TypeError: If any of the types of the arguments are wrong.
            ValueError: Tag label already exists.
            ValueError: Parent tag does not exist.
            ValueError: The hex string color inputted is invalid.

        !!! example
            ```python
            tag_name = 'Books'  # The name for our tag
            parent_name = 'Productivity'  # The desired parent tag -> this should already exist.
            color_code = '#1387c4'
            sort_type = 1  # Sort by `dueDate`
            tag_object = client.tag.builder(tag_name, parent=parent_name, color=color_code, sort=sort_type)
            ```

            ??? success "Result"
                The required fields to create a tag object are created and returned in a dictionary.

                ```python
                {'label': 'Fiction', 'color': '#1387c4', 'parent': 'books', 'sortType': 'dueDate', 'name': 'fiction'}
                ```
        """
        # Perform checks
        return self._check_fields(label, color=color, parent_label=parent, sort=sort)

    def create(
        self, label, color: str = "random", parent: str = None, sort: int = None
    ):
        """
        Creates a tag remotely. Supports single tag creation or batch tag creation.

        !!! tip
            Allows creation with a label that may normally not be allowed by `TickTick` for tags.

            Normal `TickTick` excluded characters are: \\ / " # : * ? < > | Space

        Arguments:
            label (str or list):
                **Single Tag (str)**: The desired label of the tag. Tag labels cannot be repeated.

                **Multiple Tags (list)**: A list of tag objects created using the [builder][managers.tags.TagsManager.builder] method.
            color: Hex color string. A random color will be generated if no color is specified.
            parent: The label of the parent tag if desired (include capitals in if it exists).
            sort: The desired sort type of the tag. Valid integer values are present in the [sort dictionary](tags.md#sort-dictionary). The default
                sort value will be by 'project'

        Returns:
            dict or list:
            **Single Tag (dict)**: The created tag object dictionary.

            **Multiple Tags (list)**: A list of the created tag object dictionaries.

        Raises:
            TypeError: If any of the types of the arguments are wrong.
            ValueError: Tag label already exists.
            ValueError: Parent tag does not exist.
            ValueError: The hex string color inputted is invalid.
            RuntimeError: The tag(s) could not be created.

        !!! example "Single Tag"

            === "Just A Label"
                ```python
                tag = client.tag.create('Fun')
                ```

                ??? success "Result"
                    The tag object dictionary is returned.

                    ```python
                    {'name': 'fun', 'label': 'Fun', 'sortOrder': 0, 'sortType': 'project', 'color': '#9b69f3', 'etag': '7fc8zb58'}
                    ```
                    Our tag is created.

                    ![image](https://user-images.githubusercontent.com/56806733/104658773-5bbb5500-5678-11eb-9d44-27214203d70e.png)

            === "Specify a Color"
                A random color can be generated using [generate_hex_color][helpers.hex_color.generate_hex_color].
                However, just not specifying a color will automatically generate a random color (as seen in the previous tab)
                You can always specify the color that you want.

                ```python
                tag = client.tag.create('Fun', color='#86bb6d')
                ```

                ??? success "Result"
                    The tag object dictionary is returned and our project is created with the color specified.

                    ```python
                    {'name': 'fun', 'label': 'Fun', 'sortOrder': 0, 'sortType': 'project', 'color': '#86bb6d', 'etag': '8bzzdws3'}
                    ```
                    ![image](https://user-images.githubusercontent.com/56806733/104659184-0c295900-5679-11eb-9f3c-2cd154c0500c.png)

            === "Specifying a Parent Tag"
                Tags can be nested one level. To create a tag that is nested, include the label of the parent tag.
                The parent tag should already exist.

                ```python
                tag = client.tag.create('Fun', parent='Hobbies')
                ```

                ??? success "Result"
                    The tag object dictionary is returned and our tag is created nested under the parent tag.

                    ```python
                    {'name': 'fun', 'label': 'Fun', 'sortOrder': 0, 'sortType': 'project', 'color': '#d2a6e4', 'etag': 'nauticx1', 'parent': 'hobbies'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104659785-24e63e80-567a-11eb-9a62-01ebca55e649.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104659814-33ccf100-567a-11eb-8dca-c91aea68b4c7.png)

            === "Sort Type"
                You can specify the sort type of the created tag using integer values from the [sort dictionary](#sort-dictionary).

                ```python
                tag = client.tag.create('Fun', sort=2)  # Sort by `title`
                ```

                ??? success "Result"
                    The tag object dictionary is returned and our tag has the specified sort type.

                    ```python
                    {'name': 'fun', 'label': 'Fun', 'sortOrder': 0, 'sortType': 'title', 'color': '#e7e7ba', 'etag': 'n4k3pezc'}
                    ```

                    ![image](https://user-images.githubusercontent.com/56806733/104660156-e4d38b80-567a-11eb-8c61-8fb874a515a2.png)

        !!! example "Multiple Tag Creation (batch)"
            To create multiple tags, build the tag objects first using the [builder][managers.projects.ProjectManager.builder] method. Pass
            in a list of the project objects to create them remotely.

            ```python
            parent_tag = client.tag.create('Hobbies')  # Create a parent tag.
            # We will create tag objects using builder that will be nested under the parent tag
            fun_tag = client.tag.builder('Fun', sort=2, parent='Hobbies')
            read_tag = client.tag.builder('Read', color='#d2a6e4', parent='Hobbies')
            movie_tag = client.tag.builder('Movies', parent='Hobbies')
            # Create the tags
            tag_list = [fun_tag, read_tag, movie_tag]
            created_tags = client.tag.create(tag_list)
            ```

            ??? success "Result"
                The tag object dictionaries are returned in a list.

                ```python
                [{'name': 'fun', 'label': 'Fun', 'sortOrder': 0, 'sortType': 'title', 'color': '#172d1c', 'etag': '1tceclp4', 'parent': 'hobbies'},

                {'name': 'read', 'label': 'Read', 'sortOrder': 0, 'sortType': 'project', 'color': '#d2a6e4', 'etag': 'ykdem8dg', 'parent': 'hobbies'},

                {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'project', 'color': '#94a5f8', 'etag': 'o0nifkbv', 'parent': 'hobbies'}]
                ```

                ![image](https://user-images.githubusercontent.com/56806733/104660625-cb7f0f00-567b-11eb-8649-68646870ccfa.png)
        """
        batch = False  # Bool signifying batch create or not
        if isinstance(label, list):
            # Batch tag creation triggered
            obj = label  # Assuming all correct objects
            batch = True
        else:
            if not isinstance(label, str):
                raise TypeError(
                    "Required Positional Argument Must Be A String or List of Tag Objects"
                )
            # Create a single object
            obj = self.builder(label=label, color=color, parent=parent, sort=sort)

        if not batch:
            obj = [obj]

        url = self._client.BASE_URL + "batch/tag"
        payload = {"add": obj}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()

        if not batch:
            return self._client.get_by_etag(
                self._client.parse_etag(response), search="tags"
            )
        else:
            etag = response["id2etag"]
            etag2 = list(etag.keys())  # Tag names are out of order
            labels = [x["name"] for x in obj]  # Tag names are in order
            items = [""] * len(obj)  # Create enough spots for the objects
            for tag in etag2:
                index = labels.index(tag)  # Object of the index is here
                actual_etag = etag[tag]  # Get the actual etag
                found = self._client.get_by_etag(actual_etag, search="tags")
                items[index] = found  # Place at the correct index
            if len(items) == 1:
                return items[0]
            else:
                return items

    def rename(self, old: str, new: str) -> dict:
        """
        Renames a tag.

        Arguments:
            old: Current label of the tag to be changed.
            new: Desired new label of the tag.

        Returns:
            The tag object with the updated label.

        Raises:
            TypeError: If `old` and `new` are not strings.
            ValueError: If the `old` tag label does not exist.
            ValueError: If the `new` tag label already exists.
            RuntimeError: If the renaming was unsuccessful.

        !!! example "Changing a Tag's Label"

            Pass in the current label of the tag, and the desired new label of the tag.

            ```python
            # Lets assume that we have a tag that already exists named "Movie"
            old_label = "Movie"
            new_label = "Movies"
            updated_tag = client.tag.rename(old_label, new_label)
            ```

            ??? success "Result"
                The updated tag object dictionary is returned.

                ```python
                {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'project', 'color': '#134397', 'etag': 'qer1jygy'}
                ```

                **Before**

                ![image](https://user-images.githubusercontent.com/56806733/104661255-fcac0f00-567c-11eb-9f10-69af8b50e0b4.png)

                **After**

                ![image](https://user-images.githubusercontent.com/56806733/104661299-19e0dd80-567d-11eb-825f-758d83178295.png)
        """
        # Check that both old and new are strings
        if not isinstance(old, str) or not isinstance(new, str):
            raise TypeError("Old and New Must Be Strings")

        # Make sure the old tag exists
        old = old.lower()
        # Check if the tag object exists
        obj = self._client.get_by_fields(name=old, search="tags")
        if not obj:
            raise ValueError(f"Tag '{old}' Does Not Exist To Rename")

        # Make sure the new tag does not exist
        temp_new = new.lower()
        # Check if the tag object exists
        found = self._client.get_by_fields(name=temp_new, search="tags")
        if found:
            raise ValueError(f"Name '{new}' Already Exists -> Cannot Duplicate Name")

        url = self._client.BASE_URL + "tag/rename"
        payload = {"name": obj["name"], "newName": new}
        response = self._client.http_put(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        # Response from TickTick does not return the new etag of the object, we must find it ourselves
        new_obj = self._client.get_by_fields(name=temp_new, search="tags")
        # Return the etag of the updated object
        return self._client.get_by_etag(new_obj["etag"], search="tags")

    def color(self, label: str, color: str) -> dict:
        """
        Change the color of a tag. For batch changing colors, see [update][managers.tags.TagsManager.update].

        Arguments:
            label: The label of the tag to be changed.
            color: The new desired hex color string.

        Returns:
            The updated tag dictionary object.

        Raises:
            TypeError: If `label` or `color` are not strings.
            ValueError: If the tag `label` does not exist.
            ValueError: If `color` is not a valid hex color string.
            RuntimeError: If changing the color was not successful.

        !!! example "Changing a Tag's Color"
            ```python
            # Lets assume that we have a tag named "Movies" that we want to change the color for.
            new_color = '#134397'
            movies_updated = client.tag.color('Movies', new_color)
            ```

            ??? success "Result"
                The updated tag dictionary object is returned.

                ```python
                {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'project', 'color': '#134397', 'etag': 'wwb49yfr'}
                ```

                **Before**

                ![image](https://user-images.githubusercontent.com/56806733/104661749-0eda7d00-567e-11eb-836f-3a8851bcf9a5.png)

                **After**

                ![image](https://user-images.githubusercontent.com/56806733/104661860-55c87280-567e-11eb-93b5-054fa4f1104a.png)
        """
        if not isinstance(label, str) or not isinstance(color, str):
            raise TypeError("Label and Color Must Be Strings")

        # Get the object
        label = label.lower()
        obj = self._client.get_by_fields(name=label, search="tags")
        if not obj:
            raise ValueError(f"Tag '{label}' Does Not Exist To Update")

        # Check the color
        if not check_hex_color(color):
            raise ValueError(f"Hex Color String '{color}' Is Not Valid")

        obj["color"] = color  # Set the color

        url = self._client.BASE_URL + "batch/tag"
        payload = {"update": [obj]}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        return self._client.get_by_etag(response["id2etag"][obj["name"]])

    def sorting(self, label: str, sort: int) -> dict:
        """
        Change the sort type of a tag. For batch changing sort types, see [update][managers.tags.TagsManager.update].

        Arguments:
            label: The label of the tag to be changed.
            sort: The new sort type specified by an integer 0-3. See [sort dictionary](tags.md#sort-dictionary).

        Returns:
            The updated tag dictionary object.

        Raises:
            TypeError: If `label` is not a string or if `sort` is not an int.
            ValueError: If the tag `label` does not exist.
            RuntimeError: If the updating was unsuccessful.

        !!! example "Changing the Sort Type"

            ```python
            # Lets assume that we have a tag named "Movies" with the sort type "project"
            changed_sort_type = client.tag.sorting("Movies", 1)  # Sort by 'dueDate'
            ```

            ??? success "Result"
                The updated task dictionary object is returned.

                ```python
                {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'dueDate', 'color': '#134397', 'etag': 'fflj8iy0'}
                ```

                **Before**

                ![image](https://user-images.githubusercontent.com/56806733/104663625-3f241a80-5682-11eb-93a7-73d280c59b3e.png)

                **After**

                ![image](https://user-images.githubusercontent.com/56806733/104663663-5531db00-5682-11eb-9440-5673a70840b4.png)
        """
        if not isinstance(label, str) or not isinstance(sort, int):
            raise TypeError("Label Must Be A String and Sort Must Be An Int")

        # Get the object
        label = label.lower()
        obj = self._client.get_by_fields(name=label, search="tags")
        if not obj:
            raise ValueError(f"Tag '{label}' Does Not Exist To Update")
        sort = self._sort_string_value(sort)  # Get the sort string for the value

        obj["sortType"] = sort  # set the object field

        url = self._client.BASE_URL + "batch/tag"
        payload = {"update": [obj]}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        return self._client.get_by_etag(response["id2etag"][obj["name"]])

    def nesting(self, child: str, parent: str) -> dict:
        """
        Update tag nesting. Move an already created tag to be nested underneath a parent tag - or ungroup an already
        nested tag.

        !!! warning "Nesting Tags More Than One Level Does Not Work"
            !!! example

                === "Nesting Explanation"
                    ```md
                    Parent Tag -> Level Zero
                        Child Tag 1 -> Level One: This is the most nesting that is allowed by TickTick for tags.
                            Child Tag 2 -> Level Two: Not allowed
                    ```

        Arguments:
            child: Label of the tag to become the child
            parent: Label of the tag that will become the parent.

        Returns:
            The updated tag object dictionary.

        Raises:
            TypeError: If `child` and `parent` are not strings
            ValueError: If `child` does not exist to update.
            ValueError: If `parent` does not exist.
            RuntimeError: If setting the parent was unsuccessful.

        !!! example "Nesting"

            === "Nesting A Tag"
                To nest a tag underneath another tag, pass in the labels of the child and parent.

                ```python
                # Lets assume that we have a tag named "Movies"
                # We have another tag named "Hobbies" that we want to make the parent to "Movies"
                child = "Movies"
                parent = "Hobbies"
                nesting_update = client.tag.nesting(child, parent)
                ```

                ??? success "Result"
                    The updated child tag dictionary object is returned.

                    ```python
                    {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'dueDate', 'color': '#134397', 'etag': 'ee34aft9', 'parent': 'hobbies'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104665300-da6abf00-5685-11eb-947f-889187cec008.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104665366-f706f700-5685-11eb-93eb-9316befec5fc.png)

            === "Changing The Parent Of An Already Nested Tag"
                If the tag is already nested, changing the parent is still no different.

                ```python
                # We have a tag named "Movies" that is already nested underneath "Hobbies"
                # We want to nest "Movies" underneath the tag "Fun" instead.
                child = "Movies"
                parent = "Fun"
                nesting_update = client.tag.nesting(child, parent)
                ```

                ??? success "Result"
                    The updated child tag dictionary object is returned.

                    ```python
                    {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'dueDate', 'color': '#134397', 'etag': '91qpuq71', 'parent': 'fun'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104665599-ab088200-5686-11eb-8b36-5ee873289db7.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104665821-35e97c80-5687-11eb-8098-426816970f3e.png)

            === "Un-grouping A Child Tag"
                If the tag is nested and you want to ungroup it, pass in `None` for `parent`.

                ```python
                # We have a tag named "Movies" that is nested underneath "Fun"
                # We don't want to have "Movies" nested anymore.
                child = "Movies"
                parent = None
                nesting_update = client.tag.nesting(child, parent)
                ```

                ??? success "Result"
                    The updated child tag dictionary object is returned.

                    ```python
                    {'name': 'movies', 'label': 'Movies', 'sortOrder': 0, 'sortType': 'dueDate', 'color': '#134397', 'etag': 'jcoc94p6'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104666038-be681d00-5687-11eb-8490-83c370977267.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104666080-dcce1880-5687-11eb-9ca8-5abcdb4109ba.png)
        """
        if not isinstance(child, str):
            raise TypeError("Inputs Must Be Strings")

        if parent is not None:
            if not isinstance(parent, str):
                raise TypeError("Inputs Must Be Strings")

        # Get the object
        child = child.lower()
        obj = self._client.get_by_fields(name=child, search="tags")
        if not obj:
            raise ValueError(f"Tag '{child}' Does Not Exist To Update")

        # Four Cases
        # Case 1: No Parent -> Want a Parent
        # Case 2: No Parent -> Doesn't Want a Parent
        # Case 3: Has Parent -> Wants a Different Parent
        # Case 4: Has Parent -> Doesn't Want a Parent

        # Case 1: Determine if the object has a parent
        try:
            if obj["parent"]:
                # It has a parent
                if parent is not None:  # Case 3
                    # check if the parent is already the same, if it is just return
                    if obj["parent"] == parent.lower():
                        return obj
                    else:
                        new_p = parent.lower()
                        obj["parent"] = new_p
                else:
                    new_p = obj["parent"]  # Case 4
                    obj["parent"] = ""
            elif obj["parent"] is None:
                raise ValueError("Parent Does Not Exist")

        except KeyError:
            # It does not have a parent
            if parent is not None:  # Wants a different parent
                new_p = parent.lower()  # -> Case 1
                obj["parent"] = new_p
            else:  # Doesn't want a parent -> Case 2
                return obj  # We don't have to do anything if no parent and doesn't want a parent

        # Have to find the project
        pobj = self._client.get_by_fields(name=new_p, search="tags")
        if not pobj:
            raise ValueError(f"Tag '{parent}' Does Not Exist To Set As Parent")

        url = self._client.BASE_URL + "batch/tag"
        payload = {"update": [pobj, obj]}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()
        return self._client.get_by_etag(response["id2etag"][obj["name"]], search="tags")

    def update(self, obj):
        """
        Generic update method. Supports single and batch tag update.

        !!! important
            Updating tag properties like `parent` and renaming tags must be completed through
            their respective class methods to work: [nesting][managers.tags.TagsManager.nesting]
            and [renaming][managers.tags.TagsManager.rename]. These updates use different
            endpoints to the traditional updating.

        !!! important
            You are able to batch update sorting and color of tag objects through this method. If you only
            need to update single tags, it is recommended you use the class methods: [sorting][managers.tags.TagsManager.sorting]
            and [color][managers.tags.TagsManager.color]

        !!! info
            More information on Tag Object properties [here](tags.md#example-ticktick-tag-dictionary)

        Arguments:
            obj (dict or list):
                **Single Tag (dict)**: The tag dictionary object to update.

                **Multiple Tags (list)**: The tag dictionaries to update in a list.

        Returns:
            dict or list:
            **Single Tag (dict)**: The updated tag dictionary object.

            **Multiple Tags (list)**: The updated tag dictionaries in a list.

        Raises:
            TypeError: If `obj` is not a dict or list.
            RuntimeError: If the updating was unsuccessful.

        !!! example "Updating Tags"
            === "Single Tag Update"
                Change a field directly in the task object then pass it to the method. See above
                for more information about what can actually be successfully changed through this method.

                ```python
                # Lets say we have a tag named "Fun" that we want to change the color of.
                # We can change the color by updating the field directly.
                fun_tag = client.get_by_fields(label='Fun', search='tags')  # Get the tag object
                new_color = '#d00000'
                fun_tag['color'] = new_color  # Change the color
                updated_fun_tag = client.tag.update(fun_tag)  # Pass the object to update.
                ```

                ??? success "Result"
                    The updated tag dictionary object is returned.

                    ```python
                    {'name': 'fun', 'label': 'Fun', 'sortOrder': 2199023255552, 'sortType': 'project', 'color': '#d00000', 'etag': 'i85c8ijo'}
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104669635-4aca0e00-568f-11eb-8bc6-9572a432b623.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104669824-ac8a7800-568f-11eb-93d6-ac40235bcd3f.png)

            === "Multiple Tag Update"
                Changing the fields is the same as with updating a single tag, except you will need
                to pass the objects in a list to the method.

                ```python
                # Lets update the colors for three tags: "Fun", "Hobbies", and "Productivity"
                fun_tag = client.get_by_fields(label="Fun", search='tags')
                hobbies_tag = client.get_by_fields(label="Hobbies", search='tags')
                productivity_tag = client.get_by_fields(label="Productivity", search='tags')
                fun_color_new = "#951a63"
                hobbies_color_new = "#0f8a1f"
                productivity_color_new = "#493293"
                # Change the fields directly
                fun_tag['color'] = fun_color_new
                hobbies_tag['color'] = hobbies_color_new
                productivity_tag['color'] = productivity_color_new
                # The objects must be passed in a list
                update_tag_list = [fun_tag, hobbies_tag, productivity_tag]
                updated_tags = client.tag.update(update_tag_list)
                ```

                ??? success "Result"
                    The updated task dictionary objects are returned in a list.

                    ```python
                    [{'name': 'fun', 'label': 'Fun', 'sortOrder': -1099511627776, 'sortType': 'project', 'color': '#951a63', 'etag': 'n543ajq2'},

                    {'name': 'hobbies', 'label': 'Hobbies', 'sortOrder': -549755813888, 'sortType': 'project', 'color': '#0f8a1f', 'etag': 'j4nspkg4'},

                    {'name': 'productivity', 'label': 'Productivity', 'sortOrder': 0, 'sortType': 'project', 'color': '#493293', 'etag': '34qz9bzq'}]
                    ```

                    **Before**

                    ![image](https://user-images.githubusercontent.com/56806733/104670498-cd070200-5690-11eb-9fdd-0287fa6c7e7b.png)

                    **After**

                    ![image](https://user-images.githubusercontent.com/56806733/104670531-dc864b00-5690-11eb-844a-899031335922.png)

        """
        batch = False  # Bool signifying batch create or not
        if isinstance(obj, list):
            # Batch tag creation triggered
            obj_list = obj  # Assuming all correct objects
            batch = True
        else:
            if not isinstance(obj, dict):
                raise TypeError(
                    "Required Positional Argument Must Be A Dict or List of Tag Objects"
                )

        if not batch:
            obj_list = [obj]

        url = self._client.BASE_URL + "batch/tag"
        payload = {"update": obj_list}
        response = self._client.http_post(
            url, json=payload, cookies=self._client.cookies, headers=self.headers
        )
        self._client.sync()

        if not batch:
            return self._client.get_by_etag(
                self._client.parse_etag(response), search="tags"
            )
        else:
            etag = response["id2etag"]
            etag2 = list(etag.keys())  # Tag names are out of order
            labels = [x["name"] for x in obj_list]  # Tag names are in order
            items = [""] * len(obj_list)  # Create enough spots for the objects
            for tag in etag2:
                index = labels.index(tag)  # Object of the index is here
                actual_etag = etag[tag]  # Get the actual etag
                found = self._client.get_by_etag(actual_etag, search="tags")
                items[index] = found  # Place at the correct index
            return items

    def merge(self, label, merged: str):
        """

        Merges the tasks of the passed tags into the argument `merged` and deletes all the tags except `merged`
        Args can be individual label strings, or a list of strings

        Arguments:
            label (str or list):
                **Single Tag (str)**: The label string of the tag to merge.

                **Multiple Tags (list)**: The label strings of the tags to merge in a list.
            merged: The label of the tag that will remain after the merge.

        Returns:
            dict: The tag dictionary object that remains after the merge.

        Raises:
            TypeError: If `merged` is not a str or if `label` is not a str or list.
            ValueError: If any of the labels do not exist.
            RuntimeError: If the merge could not be successfully completed.

        !!! example "Merging Tags"
            === "Merging Two Tags"
                Merging two tags requires the label of the tag that you want kept after the merge, and the
                label of the tag that will be merged.

                Lets assume that we have two tags: "Work" and "School". I want to merge the tag "School"
                into "Work". What should happen is that any tasks that are tagged "School", will be updated
                to have the tag "Work", and the "School" tag will be deleted.

                ```python
                merged_tags = client.tag.merge("School", "Work")
                ```

                ??? success "Result"
                    The tag that remains after the merge is returned.

                    ```python
                    {'name': 'work', 'label': 'Work', 'sortOrder': 2199023255552, 'sortType': 'project', 'color': '#3876E4', 'etag': 'eeh8zrup'}
                    ```

                    **Before**

                    "School" has two tasks that have it's tag.

                    ![image](https://user-images.githubusercontent.com/56806733/104680244-45c38980-56a4-11eb-968d-884160c77247.png)

                    "Work" has no tasks.

                    ![image](https://user-images.githubusercontent.com/56806733/104680366-90dd9c80-56a4-11eb-975f-5e769e9ea491.png)

                    **After**

                    "School" has been deleted. The tasks that used to be tagged with "School" are now
                    tagged with "Work".

                    ![image](https://user-images.githubusercontent.com/56806733/104680576-0c3f4e00-56a5-11eb-9536-ef3a7fcf20ec.png)

            === "Merging Three Or More Tags"
                Merging multiple tags into a single tag requires passing the labels of the tags to merge in a list.

                Lets assume that we have three tags: "Work", "School", and "Hobbies" . I want to merge the tag "School"
                and the tag "Hobbies" into "Work". What should happen is that any tasks that are tagged with "School" or "Hobbies", will be updated
                to have the tag "Work", and the "School" and "Hobbies" tags will be deleted.

                ```python
                merge_tags = ["School", "Hobbies"]
                result = client.tag.merge(merge_tags, "Work")
                ```

                ??? success "Result"
                    The tag that remains after the merge is returned.

                    ```python
                    {'name': 'work', 'label': 'Work', 'sortOrder': 2199023255552, 'sortType': 'project', 'color': '#3876E4', 'etag': 'ke23lp06'}
                    ```

                    **Before**

                    "School" has two tasks.

                    ![image](https://user-images.githubusercontent.com/56806733/104681135-7ad0db80-56a6-11eb-81dd-03e4a151cfd9.png)

                    "Hobbies" has two tasks.

                    ![image](https://user-images.githubusercontent.com/56806733/104681104-67257500-56a6-11eb-99b0-57bbb876a59e.png)

                    "Work" has one task.

                    ![image](https://user-images.githubusercontent.com/56806733/104681164-89b78e00-56a6-11eb-99a8-c85ef418d2a0.png)

                    **After**

                    "Work" has five tasks now, and the tags "School" and "Hobbies" have been deleted.

                    ![image](https://user-images.githubusercontent.com/56806733/104681239-b7043c00-56a6-11eb-9b45-5522b9c69cb0.png)
        """
        # Make sure merged is a string
        if not isinstance(merged, str):
            raise ValueError("Merged Must Be A String")

        # Make sure label is a string or list
        if not isinstance(label, str) and not isinstance(label, list):
            raise ValueError(f"Label must be a string or a list.")

        # Lowercase merged
        merged = merged.lower()
        # Make sure merged exists
        kept_obj = self._client.get_by_fields(name=merged, search="tags")
        if not kept_obj:
            raise ValueError(f"Kept Tag '{merged}' Does Not Exist To Merge")

        merge_queue = []
        # Verify all args are valid, and add them to a list
        if isinstance(label, str):
            string = label.lower()
            # Make sure it exists
            retrieved = self._client.get_by_fields(name=string, search="tags")
            if not retrieved:
                raise ValueError(f"Tag '{label}' Does Not Exist To Merge")
            merge_queue.append(retrieved)
        else:
            for (
                item
            ) in (
                label
            ):  # Loop through the items in the list and check items are a string and exist
                # Make sure the item is a string
                if not isinstance(item, str):
                    raise ValueError(f"Item '{item}' Must Be A String")
                string = item.lower()
                # Make sure it exists
                found = self._client.get_by_fields(name=string, search="tags")
                if not found:
                    raise ValueError(f"Tag '{item}' Does Not Exist To Merge")
                merge_queue.append(found)

        for labels in merge_queue:
            # Merge
            url = self._client.BASE_URL + "tag/merge"
            payload = {"name": labels["name"], "newName": kept_obj["name"]}
            self._client.http_put(
                url, json=payload, cookies=self._client.cookies, headers=self.headers
            )
        self._client.sync()

        return kept_obj

    def delete(self, label):
        """
        Delete tag(s). Supports single tag deletion and "mock" batch tag deletion.

        !!! info
            Batch deleting for tags is not supported by TickTick. However, passing in
            a list of labels to delete will "mock" batch deleting - but individual requests
            will have to be made for each deletion.

        Arguments:
            label (str or list):
                **Single Tag (str)**: The label of the tag.

                **Multiple Tags (list)**: A list of tag label strings.

        Returns:
            dict or list:
            **Single Tag (dict)**: The dictionary object of the deleted tag.

            **Multiple Tags (list)**: The dictionary objects of the deleted tags in a list.

        Raises:
            TypeError: If `label` is not a string or list.
            ValueError: If a label does not exist.
            RuntimeError: If the tag could not be deleted successfully.

        !!! example "Tag Deletion"
            === "Single Tag Deletion"
                Deleting a single tag requires passing in the label string of the tag.

                ```python
                # Lets delete a tag named "Fun"
                delete_tag = client.tag.delete("Fun")
                ```

                ??? success "Result"
                    The dictionary object of the deleted tag returned.

                    ```python
                    {'name': 'fun', 'label': 'Fun', 'sortOrder': -3298534883328, 'sortType': 'project', 'color': '#A9949E', 'etag': '32balm5l'}
                    ```

                    **Before**

                    "Fun" Tag Exists

                    ![image](https://user-images.githubusercontent.com/56806733/104668024-2c164800-568c-11eb-853e-5b7eba1f4528.png)

                    **After**

                    "Fun" Tag Does Not Exist

                    ![image](https://user-images.githubusercontent.com/56806733/104667768-ac887900-568b-11eb-9bfb-597c752e4c3b.png)

            === "Multiple Tag Deletion"
                Deleting multiple tags requires passing the label strings of the tags in a list.

                ```python
                # Lets delete tags named "Fun", "Movies", and "Hobbies"
                delete_labels = ["Fun", "Movies", "Hobbies"]
                deleted_tags = client.tag.delete(delete_labels)
                ```

                ??? success "Result"

                    The dictionary object of the deleted tags returned in a list.

                    ```python
                    [{'name': 'fun', 'label': 'Fun', 'sortOrder': -3848290697216, 'sortType': 'project', 'color': '#FFD966', 'etag': '56aa6dva'},

                    {'name': 'movies', 'label': 'Movies', 'sortOrder': -2748779069440, 'sortType': 'dueDate', 'color': '#134397', 'etag': 's0czro3e'},

                    {'name': 'hobbies', 'label': 'Hobbies', 'sortOrder': -2199023255552, 'sortType': 'project', 'color': '#ABA6B5', 'etag': 'shu2xbvq'}]
                    ```

                    **Before**

                    All three tags exist.

                    ![image](https://user-images.githubusercontent.com/56806733/104668135-61bb3100-568c-11eb-8707-314deb42cd1d.png)

                    **After**

                    All three tags don't exist.

                    ![image](https://user-images.githubusercontent.com/56806733/104668185-7b5c7880-568c-11eb-8da0-aaee68d53500.png)

        """
        # Determine if the tag exists
        if not isinstance(label, str) and not isinstance(label, list):
            raise TypeError("Label Must Be A String or List Of Strings")

        url = self._client.BASE_URL + "tag"
        if isinstance(label, str):
            label = [label]  # If a singular string we are going to add it to a list

        objects = []
        for lbl in label:
            if not isinstance(lbl, str):
                raise TypeError(f"'{lbl}' Must Be A String")
            lbl = lbl.lower()
            tag_obj = self._client.get_by_fields(
                name=lbl, search="tags"
            )  # Get the tag object
            if not tag_obj:
                raise ValueError(f"Tag '{lbl}' Does Not Exist To Delete")
            # We can assume that only one tag has the name
            params = {"name": tag_obj["name"]}
            response = self._client.http_delete(
                url, params=params, cookies=self._client.cookies, headers=self.headers
            )
            # Find the tag in the tags list and delete it, then return the deleted object
            objects.append(
                self._client.delete_from_local_state(
                    search="tags", etag=tag_obj["etag"]
                )
            )
        self._client.sync()
        if len(objects) == 1:
            return objects[0]
        else:
            return objects



#hierarchy.py
#Provides functions and classes for organizing text into fields in a console.
# Convenience Functions:
#
#  hier_out(...) <-- Takes a nested list and generates an ASCII tree
#
# Classes:
#
#  Hierarchy(...) <-- Represents objects that associate in a hierarchy
#
# One-off niche helper functions:
#
#   join_contiguous(...) <-- Merges consecutive, nested lists
#   list_depth(...) <-- returns the depth of nesting as an integer
#
#
#Rob O.
#Last Updated: 8/25/2014

import weakref
import deal


class HierarchyError(Exception):
    """Indicates an exception emanating from the Hierarchy class."""

class Hierarchy(object):
    """Initializer is an element name, or a hierarchically nested list.

Objects that subclass Hierarchy will be able to:

* Have children via the self.insert(...)
* Change parentage by assigning to the self.parent attribute
* Reorganize their children with the self.permute_children(...) method

A "hierarchically nested list" is converted into a hierarchy as follows:

* The first string in the list is the element's name.
* If the list has no string element, its name is derived from its parent's name.
* If the list has no parent (i.e., if the element is the root), its default
    name is 'root'.
* Remaining non-list elements are passed to the initializer so they can be
    used by Sub-Classes (the base class ignores the remaining elements).
* Remaining list elements are recursively parsed following the above criteria.

NOTE: The parser is flexible enough to handle any nested list, although the
      result may need tweaking if the original list was not properly structured.
      The hier_out(...) function will give you a visual indication of how the
      Hierarchy class will interpret any given nested list.

EXAMPLES:

>>> print Hierarchy('A hierarchical singleton')
+------------------------+
|A hierarchical singleton|
+------------------------+
>>> h = Hierarchy('root')
>>> h.insert('child')
>>> print h
+-------+
| root  |
|+-----+|
||child||
|+-----+|
+-------+
>>> print Hierarchy(['root',['child 1',['grand-child']],['child 2']])
+------------------------+
|          root          |
|+-------------++-------+|
||   child 1   ||child 2||
||+-----------+|+-------+|
|||grand-child||         |
||+-----------+|         |
|+-------------+         |
+------------------------+
"""
    @deal.pure
    def __init__(self, initializer='root', parent=None, *args): #Remove *args?


        self.name = None
        self._sisters = []
        self._daughters = []
        self._parent = None
        self._origin = self

        #Attempt to construct an arbitrarily nested hierarchy
        if isinstance(initializer, (list, tuple, dict)):

            #this recursive function does all the work
            temp = self._parse_list(initializer)

            if temp._parent == self:
                temp._parent = None

            for k in temp.__dict__.keys():
                setattr(self, k, getattr(temp, k))

        #construct a hierarchical singleton
        #which can be subsequently attached to an existing hierarchy
        else:
            self.name = initializer

            if parent:
                self.parent = parent

    @deal.has()
    @deal.raises(HierarchyError, ValueError)
    def __getitem__(self, name):
        """Experimenting with special syntax...
A string index with the '=' symbol will be broken into two parts.
Example:
'name=super man' will return the first element with self.name == 'super man'.
'ID=99701' will return the first element with self.ID == 99701.

NOTE: Nested attributes are not allowed.  Only primitives and built-ins should
      be used in this way."""

        if isinstance(name, int):
            return self._daughters[name]

        if name.count("=") > 1:
            raise HierarchyError("'=' can appear only once per index specifier.")

        if '=' in name:
            attribute, value = name.split('=')
            attribute  = attribute.strip()
            value = value.strip()

            #return self._traverse(self.origin, value, attribute)
            return self._traverse(self, value, attribute)
        else:

            #the commented out return statement will return the first matching
            #element from ANYWHERE IN THE TREE...(the above return statment too)
            #As is, self.__get__item(...) only searches daughters of self.
            #return self._traverse(self.origin, value, attribute)
            return self._traverse(self, name)
        raise ValueError("{} not found in hierarchy.".format(name))

    @deal.has()
    @deal.raises(HierarchyError, ValueError)
    def __setitem__(self, name, value):

        if "ID={}".format(value.ID) in self:
            raise HierarchyError("Cannot assign objects already in the hierarchy.")

        if isinstance(value, Hierarchy):

            if self[name]:

                if value._parent:
                    vp = value._parent
                    vp._daughters.remove(value)
                    for daughter in vp._daughters:
                        daughter._sisters.remove(value)

                target = self[name]
                tp = target._parent
                tp._daughters.remove(target)
                for daughter in tp._daughters:
                    daughter._sisters.remove(target)

                value.parent = tp


            else:
                raise ValueError("'{}' was not found.".format(name))
        else:
            raise HierarchyError("{} must be a hierarchical type.".format(value))

    @deal.pure
    def __len__(self):
        if self._daughters:
            return sum([len(daughter) for daughter in self._daughters])+1
        else:
            return 1

    @deal.pure
    def __repr__(self):
        class_Name = str(self.__class__)
        dot_Index = class_Name.rindex('.')
        class_Name = class_Name[dot_Index+1:-2]
        #class_Name = re.search(r"\.([A-Za-z_]+)'>$", class_Name)[0]

        return "{}('{}')".format(class_Name,self.name)

    @deal.pure
    def __str__(self):
        return self.render()

    @deal.pure
    def __contains__(self, value):
        try:
            self[value]
            return True
        except ValueError:
            return False

    @property  # type: ignore[misc]
    @deal.pure
    def ID(self):
        """Read only."""
        return hash(self)

    @property  # type: ignore[misc]
    @deal.pure
    def sisters(self):
        """Read only."""
        return self._sisters

    @property  # type: ignore[misc]
    @deal.pure
    def daughters(self):
        """Read only."""
        return self._daughters
    @property  # type: ignore[misc]
    @deal.pure
    def origin(self):
        """Read only."""
        return self._origin

    @property  # type: ignore[misc]
    @deal.pure
    def parent(self):
        return self._parent

    @deal.pure
    @parent.setter
    def parent(self, target):
        """Removes self from current parent and makes target self's new parent.
Also cleans up secondary references, lateral references, etc."""

        try:
            self._traverse(self, target.ID, 'ID')
            raise HierarchyError("Cannot make a child element a parent.")
        except ValueError:
            pass

        if self._parent:
            old_Parent = self._parent
            old_Parent._daughters.remove(self)
            for daughter in old_Parent._daughters:
                daughter._sisters.remove(self)

        self._sisters = target._daughters[:]

        if target._daughters:
            for daughter in target._daughters:
                daughter._sisters.append(self)


        target._daughters.append(self)

        self._parent = target
        self._propagate(self, "_origin", target.origin)

    @deal.pure
    def _copy(self, depth = -1):
        """The recursive part of the self.copy() method."""
        temp = self.__class__(self.name)

        temp.__dict__ = self.__dict__.copy()

        if depth == 0:
            temp._daughters = []

        if depth > 0 or depth == -1:

            if depth > 0:
                depth -= 1

            #daughters are cached and the list is emptied so that copies
            #can be re-inserted individually using copy.parent assignment
            temp_daughters = [daughter.copy(depth) for daughter in temp._daughters]
            temp._daughters = []
            for daughter in temp_daughters:
                daughter._sisters = temp_daughters[:].remove(daughter)
                daughter.parent = temp

        return temp

    @deal.pure
    def _propagate(self, root, attribute, value):
        """Assign value to root.attribute and all of root's daughter.attribute."""

        setattr(root, attribute, value)

        for daughter in root._daughters:
            self._propagate(daughter, attribute, value)

    @deal.has()
    @deal.raises(ValueError)
    def _traverse(self, root, target, attribute = "name"):
        """Search the hierarchy from root->leaves for an element with target attribute."""
        prop = getattr(root, attribute, None)
        if prop:
            if prop == prop.__class__(target):
                return root
            else:
                for daughter in root._daughters:
                    temp = self._traverse(daughter, target, attribute)
                    if temp:
                        return temp

        if root==self:
            raise ValueError("{} not found in the hierarchy.".format(target))

    @deal.pure
    def _parse_list(self, nested_List):
        """Recursively turn a nested list/dictionary into a tree of Hierarchy objects."""

        # Handle the empty edge case
        if not nested_List:
            nested_list = ["root"]
        # Convert dictionary into a (str, list) tuple corresponding to (key, value)
        elif isinstance(nested_List, dict):
            nested_list = ["dict"]+[[str(k)+":->", (v if isinstance(v, (list, dict)) else [v])] for k, v in nested_List.items()]
        else:
            nested_list = nested_List[:]


        sub_Lists = [elem for elem in nested_list if isinstance(elem, (list, tuple, dict))]
        extra_Args = [elem for elem in nested_list if not isinstance(elem, (list, tuple, dict))]

        temp = self.__class__(", ".join([str(elem) for elem in extra_Args]), self, *extra_Args)

        daughters = [temp._parse_list(daughter) for daughter in sub_Lists]

        temp._daughters = daughters

        for daughter in daughters:
            daughter._parent = temp

        return temp

    @deal.pure
    def _list_depth(self, root):
        if root._daughters:
            return max([self._list_depth(elem) for elem in root.daughters])+1
        else:
            return 1

    @deal.has('import')
    @deal.safe
    def _frame_down(self, root, attributes=[]):
        from textframe import frame, parallelize
        contents = ""
        for attr in attributes:
            line = "{}.{}: {}".format(root.name,
                                      attr,
                                      getattr(root, attr, "Error"))
            contents += line + "\n"

        if contents:
            #strip trailing new line
            contents = contents[:-1]

        if not attributes:
            contents = root.name
        if root._daughters:
            daughter_boxes = [self._frame_down(elem, attributes=attributes) for elem in root._daughters]
            return frame(contents+("\n" if contents else "")+\
                         parallelize(daughter_boxes), hJust="c")
        else:
            return frame(contents, hJust='c')

    @deal.has('import')
    @deal.safe
    def _stringify(self, attributes=[], depth = 0):
        """Recursively lay self out along one line for each level of nesting.
TODO: Make indent and grouper variables to customize output style."""

        from text import parallelize

        if 'name' in attributes:
            attributes.remove('name')

        head = "\n"*depth + self.name
        attrs = ", ".join([repr(getattr(self, attr, "Error")) for attr in attributes])
        if attrs:
            head += "({})".format(attrs)

        foot = "\n"*depth + ">"

        if self._daughters:
            head += "<"

            canvas = [head]
            for daughter in self._daughters:
                canvas.append(daughter._stringify(attributes = attributes,
                                                  depth = depth+1))

                if self._daughters.index(daughter) != len(self._daughters)-1:
                    canvas.append("\n"*(depth+1) + ", ")
            canvas.append(foot)

            return parallelize(canvas)

        else:
            return head

    @deal.pure
    def _verticalize(self, attributes = [], depth = 0):
        """Recursively lay self out on lines.
TODO: Make indent and grouper variables to customize output style."""

        if 'name' in attributes:
            attributes.remove('name')

        attrs = ", ".join([repr(getattr(self, attr, "Error")) for attr in attributes])
        if attrs:
            attrs = "({})".format(attrs)

        if self._daughters:
            base_String = "  "*depth + str(self.name)+ attrs +"\n"+ "  "*depth + "<\n"
            for daughter in self._daughters:
                base_String += daughter._verticalize(attributes= attributes,
                                                     depth = depth+1) + "\n"

            base_String += "  "*depth + ">"

            return base_String

        else:
            return "  "*depth+str(self.name)+attrs

    @deal.has('import')
    @deal.safe
    def render(self, attributes=[], mode = "frames"):
        """Display the hierarchy visually.  Attributes is a list of attributes.

3 supported rendering modes:
MODE       - VALID STRING CODES
----------------------
frames     - 'frames', 'f'
vertical   - 'vertical', 'v'
horizontal - 'horizontal', 'h'"""

        try:
            from text import frame, parallelize
        except ImportError:
            mode = "v"

        #if my text module is available, use it to frame the hierarchy
        if mode in ("frames", "f"):
            return self._frame_down(self, attributes = attributes)

        #horizontal mode also requires my text module
        elif mode in ("horizontal", "h"):
            return self._stringify(attributes = attributes, depth = 0)

        #this is a simple fall-back representation
        elif mode in ("vertical", "v"):
            return self._verticalize(attributes = attributes, depth=0)

        return self._frame_down(self, attributes = attributes)

    @deal.pure
    def insert(self, *args):
        """If first argument is hierarchical, it is simply added as a child of self.
Otherwise, arguments are taken and passed to self.__init__(...) to generate
    a hierarchical object, which is added as a child of self."""

        if isinstance(args[0], Hierarchy):
            args[0].parent = self
        else:
            temp = self.__class__(*args)
            temp.parent = self

    @deal.pure
    def copy(self, depth=-1):
        """Return a copy of self including all children up to depth.
If depth == -1, copying is exhaustive.
If depth == 0, no children are copied (only self).
NOTE: Element ID's are based on location in memory.  (The copy will have a
different self.ID value than the original.)"""

        temp = self._copy(depth)
        temp._parent = None
        temp._propagate(temp, "_origin", temp)
        return temp

    @deal.has()
    @deal.raises(AttributeError)
    def set_precedence(self, index):
        """Move this element to the index'th position among its siblings."""
        if self._parent:
            self._parent._daughters.remove(self)
            self._parent._daughters.insert(index,self)

        else:
            raise AttributeError("Root element has no index.")

    @deal.has()
    @deal.raises(HierarchyError)
    def permute_daughters(self, permutation):
        """Change the order of self's daughters to match that of permutation.
Examples:
The [0, 1, 2] permutation is the identity permutation.
The [2, 1, 0] permutation is equivalent to using reversed.
The [0,2,4,...,1,3,5,...] permutation organizes daughters by parity."""

        if len(permutation) != len(self._daughters):
            raise HierarchyError("Permutation must have exactly 1 distinct index for each daughter.")

        if sorted(permutation)!=range(len(permutation)):
            raise HierarchyError("Permutation indeces must range from 0 to n in increments of 1.")

        #This code allows greater flexibility in the permutation argument
        #Any iterable with a distinct ordering will work
        #But that seemed like a good hiding place for bugs, so its out for now.
        #permutation = [permutation.index(elem) for elem in sorted(permutation)]

        for elem, index in zip(self._daughters[:], permutation):
            elem.set_precedence(index)

    @deal.pure
    def transfer(self):
        pass


@deal.pure
def list_depth(nestedList):
    """A non-list object has 0 depth.  A simple list has a depth of 1.  Etc."""

    if isinstance(nestedList, (list, tuple)):
        return max([list_depth(elem) for elem in nestedList])+1
    else:
        return 0

@deal.pure
def join_contiguous(theList, depth = -1):
    """Joins 2 or more consecutive containers in theList, up to specified depth.
If depth == -1, ALL levels of nesting are affected.
If depth == 0, the function does nothing.
If depth == 1, once-nested lists are affected.
If depth == 2, twice-nested lists are also affected.
...etc..."""

    if depth > 0 or depth == -1:

        i = 0
        #This loop cycles through the outer-most list, looking for 2 consecutive
        #list or tuple elements.  Those 2 elements are replaced by the union.
        #The union is, itself, a list so that if there is a 3rd consecutive
        #list or tuple, it will be caught during the loop's, next iteration, etc
        while i + 1 < len(theList):
            if isinstance(theList[i], (tuple, list)):
                if isinstance(theList[i+1], (tuple, list)):
                    temp = list(theList[i]) + list(theList[i+1])
                    theList.insert(i, temp)
                    del theList[i+1]
                    del theList[i+1]
                    continue
            i+=1

        if depth == -1:
            depth += 1

        #This loop goes through the list, replacing each nested list with
        #a copy that has, itself, been conjoined by a recursive call of
        #this function
        for subList in theList[:]:
            if isinstance(subList, (list, tuple)):
                n = theList.index(subList)
                theList.remove(subList)
                theList.insert(n, join_contiguous(subList, depth-1))
    return theList

@deal.pure
def hier_out(hierarchical_Object, mode = 'f'):
    """Returns a string representation of the passed object.

EXAMPLES:
>>> print hier_out(['parent',['child']])
+-------+
|parent |
|+-----+|
||child||
|+-----+|
+-------+
>>> print hier_out(['parent',['child',['grand-child']]], mode = 'v')
parent
<
  child
  <
    grand-child
  >
>
>>> print hier_out(['root',['1',['a'],['b']],['2',['i'],['ii']]], mode = 'h')
root<                 >
     1<    >, 2<     >
       a, b     i, ii
"""
    try:
        return hierarchical_Object.render(mode= mode)
    except AttributeError:
        return Hierarchy(hierarchical_Object).render(mode= mode)

if __name__ == "__main__":
    import unittest, doctest
    from testhierarchy import *

    #A simple instance for debugging purposes
    h = Hierarchy(['r',['t'],['b']])

    doctest.testmod()
    unittest.main()



import json
import errno
import logging
import deal

log = logging.getLogger(__name__)


class CacheHandler:
    """
    Handles the caching for TickTick auth tokens to disk
    """

    @deal.pure
    def __init__(self,
                 path):
        """
        Initializes the path of the cache
        :param path:
        """
        # set path
        self.path = path

    @deal.pure
    def get_cached_token(self):
        """
        Retrieves the cached token - and raises an exception if it doesn't work
        :return:
        """
        access_token_info = None
        try:
            f = open(self.path)
            access_token_info_full_string = f.read()
            f.close()
            access_token_info = json.loads(access_token_info_full_string)

        except IOError as error_occurred:
            if error_occurred == errno.ENOENT:
                log.debug(f"No cache exists at: {self.path}")
            else:
                log.warning(f"Cache could not be read at: {self.path}")

        return access_token_info

    @deal.pure
    def write_token_to_cache(self, access_token_info):
        """
        Writes the token to cache
        :return:
        """
        try:
            f = open(self.path, "w")
            f.write(json.dumps(access_token_info))
            f.close()

        except IOError:
            log.warning(f"Cache could not be written to at: {self.path}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '..'}

packages = \
['loggers',
 'parser',
 'rob',
 'rob.loggers',
 'rob.parser',
 'rob.ticktick',
 'rob.ticktick.helpers',
 'rob.ticktick.managers',
 'ticktick',
 'ticktick.helpers',
 'ticktick.managers']

package_data = \
{'': ['*'],
 'rob': ['.hypothesis/unicode_data/13.0.0/*',
         '.pytest_cache/*',
         '.pytest_cache/v/cache/*',
         'config/*',
         'dist/*',
         'misc/misc 2022/*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'deal>=4.23.2,<5.0.0',
 'docopt>=0.6.2,<0.7.0',
 'fire>=0.5.0,<0.6.0',
 'openai>=0.15.0,<0.16.0',
 'praw>=7.5.0,<8.0.0',
 'pydub==0.25.1',
 'pytaglib==1.5.0',
 'redmail>=0.4.0,<0.5.0',
 'rich>=11.2.0,<12.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer[all]>=0.7.0,<0.8.0']

setup_kwargs = {
    'name': 'robolson',
    'version': '0.3.71',
    'description': 'My collection of utilty scripts',
    'long_description': None,
    'author': 'Rob L Olson',
    'author_email': 'rob.louis@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import argparse
import sys

# Create the parser
email_parser = argparse.ArgumentParser(
    prog="py -m rob.email",
    allow_abbrev=True,
    add_help=True,
    description="Shoot a quick email",
    epilog="(C) Rob",
)

# Add the arguments
email_parser.add_argument(
    "target",
    nargs="?",
    action="store",
    default=False,
    type=str,
    help="Email Recipient",
)

email_parser.add_argument(
    "content", nargs="*", action="store", type=str, help="Email Text"
)

# Add SUBPARSER
subparsers = email_parser.add_subparsers(help="manage contacts", required=False)

# create the SUBPARSER
contact_parser = subparsers.add_parser(
    "add",
    help="Add a {name:email} pair to the quick contact dictionary.",
    exit_on_error=False,
)

add_contact_group = contact_parser.add_argument_group("contact")

add_contact_group.add_argument(
    "name", nargs=1, type=str, help="Contact's name (used as CLI shortcut)"
)
add_contact_group.add_argument("email", nargs=1, type=str, help="Contact's email")

# create the list SUBPARSER
list_parser = subparsers.add_parser("list", help="List all quick contacts.")
# parser_a.add_argument('bar', type=int, help='bar help')
